#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any

from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.widget import label, setfont
from xpra.platform.gui import force_focus
from xpra.os_util import gi_import
from xpra.exit_codes import ExitValue
from xpra.common import noop
from xpra.log import (
    Logger, CATEGORY_INFO, STRUCT_KNOWN_FILTERS,
    enable_color, enable_debug_for, disable_debug_for, debug_enabled_categories,
)

log = Logger("util")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")


def make_category_widgets(groups: dict[str, dict[str, str]], enabled: set[str], sensitive=True,
                          toggled=noop) -> tuple[list[Any], dict[str, Any]]:
    expanders = []
    widgets = {}

    def eact(expander, *_args):
        expanded = expander.get_expanded()
        if expanded:
            return
        for exp in expanders:
            if exp != expander:
                exp.set_expanded(False)

    for group, categories in groups.items():
        exp = Gtk.Expander(label=group)
        exp.connect("activate", eact)
        setfont(exp, "sans 16")
        grid = Gtk.Grid()
        grid.set_row_homogeneous(True)
        grid.set_column_homogeneous(False)
        exp.add(grid)
        row = 0
        for category, descr in categories.items():
            cb = Gtk.CheckButton(label=category)
            setfont(cb, "courier 14")
            cb.set_active(category in enabled)
            cb.connect("toggled", toggled, category)
            cb.set_sensitive(sensitive)
            grid.attach(cb, 0, row, 1, 1)
            descr = CATEGORY_INFO.get(category, "")
            if descr:
                lbl = label(descr, font="sans 14")
                lbl.set_halign(Gtk.Align.START)
                lbl.set_margin_start(32)
                grid.attach(lbl, 1, row, 1, 1)
            row += 1
            widgets[category] = cb
        expanders.append(exp)
    return expanders, widgets


class DebugConfig:

    def __init__(self, text="Configure Client Debug Categories",
                 groups=STRUCT_KNOWN_FILTERS, enabled=debug_enabled_categories,
                 enable=enable_debug_for, disable=disable_debug_for):
        self.text = text
        self.window: Gtk.Window | None = None
        self.groups = groups
        self.enabled = enabled
        self.enable = enable
        self.disable = disable
        self.is_closed = False
        self.setup_window()

    def setup_window(self) -> None:
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.close)
        self.window.set_default_size(400, 300)
        self.window.set_title("Xpra Debug Config")

        def window_deleted(*_args):
            self.is_closed = True
            self.hide()

        add_close_accel(self.window, window_deleted)
        self.window.connect("delete_event", window_deleted)

        icon_pixbuf = get_icon_pixbuf("bugs.png")
        if icon_pixbuf:
            self.window.set_icon(icon_pixbuf)
        self.window.set_position(Gtk.WindowPosition.CENTER)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(15)

        vbox.add(label(self.text, font="Sans 16"))

        for widget in make_category_widgets(self.groups, self.enabled, True, self.category_toggled)[0]:
            vbox.pack_start(widget, True, True, 0)

        def accel_close(*_args) -> None:
            self.close()

        add_close_accel(self.window, accel_close)
        vbox.show_all()
        self.window.add(vbox)

    def category_toggled(self, btn, category: str) -> None:
        active = btn.get_active()
        log("category_toggled(%s, %s) active=%s", btn, category, active)
        if active:
            self.enable(category)
        else:
            self.disable(category)

    def show(self) -> None:
        log("show()")
        if not self.window:
            self.setup_window()
        force_focus()
        self.window.show_all()
        self.window.present()

    def hide(self) -> None:
        log("hide()")
        if self.window:
            self.window.hide()

    def close(self, *_args) -> bool:
        self.hide_window()
        return True

    def hide_window(self) -> None:
        log("hide_window()")
        if self.window:
            self.hide()
            self.window = None

    def destroy(self, *args) -> None:
        log("destroy%s", args)
        if self.window:
            self.window.destroy()
            self.window = None

    @staticmethod
    def run() -> ExitValue:
        log("run()")
        Gtk.main()
        log("run() Gtk.main done")
        return 0

    def quit(self, *args) -> None:
        log("quit%s", args)
        self.hide_window()
        Gtk.main_quit()


def main(argv=()) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.platform.gui import init, set_default_icon
    from xpra.gtk.util import init_display_source
    with program_context("Xpra-Debug-Config", "Xpra Debug Config"):
        enable_color()
        init_display_source(False)
        set_default_icon("bugs.png")
        init()

        if "-v" in argv:
            enable_debug_for("util")

        from xpra.util.glib import register_os_signals
        app = DebugConfig("Configure Debug Categories",
                          STRUCT_KNOWN_FILTERS, debug_enabled_categories,
                          enable_debug_for, disable_debug_for)
        app.close = app.quit
        register_os_signals(app.quit, "Debug Config")
        try:
            from xpra.platform.gui import ready as gui_ready
            gui_ready()
            app.show()
            app.run()
        except KeyboardInterrupt:
            pass
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform import platform_import
from xpra.client.base.stub import StubClientMixin

# default:
PlatformClient: type | None = StubClientMixin

platform_import(globals(), "client", True,
                "PlatformClient")



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def get_server_base_class() -> type:
    from xpra.server import features
    # disable a bunch of things we're not using:
    features.power = features.suspend = features.idle = False
    # features.ping = auto
    features.bandwidth = features.control = False
    # features.debug = auto
    features.file = features.printer = False
    # features.mmap = auto
    features.logging = features.http = features.shell = features.ssh = features.webcam = False
    # features.dbus = auto
    features.display = False
    features.notification = features.clipboard = features.keyboard = features.pointer = False
    features.audio = features.pulseaudio = False
    features.encoding = features.cursor = features.window = False
    features.command = True

    from xpra.server.base import ServerBase
    return ServerBase



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.common import noop
from xpra.clipboard.common import env_timeout
from xpra.clipboard.core import ClipboardProtocolHelperCore
from xpra.clipboard.proxy import ClipboardProxyCore
from xpra.util.str_fn import Ellipsizer
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("clipboard")

REMOTE_TIMEOUT = env_timeout("REMOTE", 2500)


class ClipboardTimeoutHelper(ClipboardProtocolHelperCore):
    # a clipboard superclass that handles timeouts

    def __init__(self, send_packet_cb: Callable, progress_cb=noop, **kwargs):
        super().__init__(send_packet_cb, progress_cb, **kwargs)
        self._clipboard_outstanding_requests: dict[int, tuple[int, str, str]] = {}

    def cleanup(self) -> None:
        # reply to outstanding requests with "no data":
        for request_id in tuple(self._clipboard_outstanding_requests.keys()):
            self._clipboard_got_contents(request_id)
        self._clipboard_outstanding_requests = {}
        super().cleanup()

    ############################################################################
    # network methods for communicating with the remote clipboard:
    ############################################################################

    def _send_clipboard_request_handler(self, proxy: ClipboardProxyCore, selection: str, target: str):
        log("send_clipboard_request_handler%s", (proxy, selection, target))
        request_id = self._clipboard_request_counter
        self._clipboard_request_counter += 1
        remote = self.local_to_remote(selection)
        log("send_clipboard_request %s to %s, id=%s", selection, remote, request_id)
        timer = GLib.timeout_add(REMOTE_TIMEOUT, self.timeout_request, request_id)
        self._clipboard_outstanding_requests[request_id] = (timer, selection, target)
        self.progress()
        self.send("clipboard-request", request_id, remote, target)

    def timeout_request(self, request_id: int) -> None:
        try:
            selection, target = self._clipboard_outstanding_requests.pop(request_id)[1:]
        except KeyError:
            log.warn("Warning: clipboard request id %i not found", request_id)
            return
        finally:
            self.progress()
        log.warn("Warning: remote clipboard request timed out")
        log.warn(" request id %i, selection=%s, target=%s", request_id, selection, target)
        proxy = self._get_proxy(selection)
        if proxy:
            proxy.got_contents(target)

    def _clipboard_got_contents(self, request_id: int, dtype: str = "", dformat: int = 0, data=None) -> None:
        try:
            timer, selection, target = self._clipboard_outstanding_requests.pop(request_id)
        except KeyError:
            log.warn("Warning: request id %i not found", request_id)
            log.warn(" already timed out or duplicate reply")
            return
        finally:
            self.progress()
        GLib.source_remove(timer)
        proxy = self._get_proxy(selection)
        log("clipboard got contents%s: proxy=%s for selection=%s",
            (request_id, dtype, dformat, Ellipsizer(data)), proxy, selection)
        if data and isinstance(data, memoryview):
            data = bytes(data)
        if proxy:
            proxy.got_contents(target, dtype, dformat, data)

    def client_reset(self) -> None:
        super().client_reset()
        # timeout all pending requests
        cor = self._clipboard_outstanding_requests
        if cor:
            log.info("cancelling %i clipboard requests", len(cor))
            self._clipboard_outstanding_requests = {}
            for request_id in cor:
                self._clipboard_got_contents(request_id)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.net.glib_handler import GLibPacketHandler
from xpra.client.base.stub import StubClientMixin
from xpra.net.common import Packet


class GLibClient(GLibPacketHandler, StubClientMixin):
    """
    Ensures that a `GLibPacketHandler` can be used as a regular client mixin module,
    by simply extending `StubClientMixin` which provides default methods.

    Also, the client packet handlers don't need the `proto` argument,
    so `call_packet_handler` is overriden so we can skip sending it.
    """

    def call_packet_handler(self, handler: Callable[[Packet], []], proto, packet: Packet) -> None:
        handler(packet)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.util.objects import typedict
from xpra.os_util import gi_import
from xpra.server import ServerExitMode
from xpra.server.subsystem.stub import StubServerMixin
from xpra.log import Logger

log = Logger("timeout")

GLib = gi_import("GLib")


class IdleTimeoutServer(StubServerMixin):

    def __init__(self):
        StubServerMixin.__init__(self)
        self.server_idle_timeout = 0
        self.server_idle_timer = 0

    def init(self, opts) -> None:
        self.server_idle_timeout = opts.server_idle_timeout

    def threaded_setup(self) -> None:
        self.schedule_server_timeout()

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        self.cancel_server_timeout()

    def last_client_exited(self) -> None:
        self.schedule_server_timeout()

    def cancel_server_timeout(self) -> None:
        log("cancel_server_timeout() timer=%s", self.server_idle_timer)
        if self.server_idle_timeout <= 0:
            return
        if self.server_idle_timer:
            GLib.source_remove(self.server_idle_timer)
            self.server_idle_timer = 0

    def schedule_server_timeout(self) -> None:
        log("schedule_server_timeout() server_idle_timeout=%s", self.server_idle_timeout)
        self.cancel_server_timeout()
        self.server_idle_timer = GLib.timeout_add(self.server_idle_timeout * 1000, self.server_idle_timedout)

    def server_idle_timedout(self) -> None:
        log.info("No valid client connections for %s seconds, exiting the server", self.server_idle_timeout)
        self.clean_quit(ServerExitMode.NORMAL)

    def cleanup(self) -> None:
        self.cancel_server_timeout()

    def get_info(self, _proto) -> dict[str, Any]:
        return {
            "idle-timeout": int(self.server_idle_timeout),
        }



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from typing import Any

from xpra.exit_codes import ExitCode
from xpra.net.ssl.common import (
    get_ssl_logger,
    SSL_VERIFY_WRONG_HOST, SSL_VERIFY_SELF_SIGNED, SSL_VERIFY_IP_MISMATCH,
    SSL_VERIFY_HOSTNAME_MISMATCH, SSL_VERIFY_CODES, SSLVerifyFailure, CERT_FILENAME, SSL_CERT_FILENAME,
)
from xpra.net.ssl.file import (
    find_ssl_cert, load_ssl_options,
    save_ssl_options, find_ssl_config_file, save_ssl_config_file,
)
from xpra.net.ssl.parsing import (
    parse_ssl_options_mask, parse_ssl_verify_mask, parse_ssl_protocol, parse_ssl_verify_mode,
)
from xpra.scripts.config import InitExit, InitException
from xpra.util.env import envbool
from xpra.util.parsing import parse_encoded_bin_data
from xpra.util.str_fn import print_nested_dict, Ellipsizer

SSL_RETRY = envbool("XPRA_SSL_RETRY", True)


def ssl_wrap_socket(sock, **kwargs):
    context, wrap_kwargs = get_ssl_wrap_socket_context(**kwargs)
    log = get_ssl_logger()
    log("ssl_wrap_socket(%s, %s) context=%s, wrap_kwargs=%s", sock, kwargs, context, wrap_kwargs)
    return do_wrap_socket(sock, context, **wrap_kwargs)


def log_ssl_info(ssl_sock) -> None:
    log = get_ssl_logger()
    log("server_hostname=%s", ssl_sock.server_hostname)
    cipher = ssl_sock.cipher()
    if cipher:
        log.info(" %s, %s bits", cipher[0], cipher[2])
    try:
        cert = ssl_sock.getpeercert()
    except ValueError:
        pass
    else:
        if cert:
            log("certificate:")
            print_nested_dict(ssl_sock.getpeercert(), prefix=" ", print_fn=log)


def ssl_handshake(ssl_sock) -> None:
    log = get_ssl_logger()
    try:
        ssl_sock.do_handshake(True)
        log.info("SSL handshake complete, %s", ssl_sock.version())
        log_ssl_info(ssl_sock)
    except Exception as e:
        log("do_handshake", exc_info=True)
        log_ssl_info(ssl_sock)
        import ssl
        ssleof_error = getattr(ssl, "SSLEOFError", None)
        if ssleof_error and isinstance(e, ssleof_error):
            return
        status = ExitCode.SSL_FAILURE
        ssl_cert_verification_error = getattr(ssl, "SSLCertVerificationError", None)
        if ssl_cert_verification_error and isinstance(e, ssl_cert_verification_error):
            verify_code = getattr(e, "verify_code", 0)
            log("verify_code=%s", SSL_VERIFY_CODES.get(verify_code, verify_code))
            try:
                msg = getattr(e, "verify_message") or (e.args[1].split(":", 2)[2])
            except (ValueError, IndexError):
                msg = str(e)
            status = ExitCode.SSL_CERTIFICATE_VERIFY_FAILURE
            log("host failed SSL verification: %s", msg)
            raise SSLVerifyFailure(status, msg, verify_code, ssl_sock) from None
        raise InitExit(status, f"SSL handshake failed: {e}") from None


def get_ssl_wrap_socket_context(cert="", key="", key_password="", ca_certs="", ca_data="",
                                protocol: str = "TLS",
                                client_verify_mode: str = "optional", server_verify_mode: str = "required",
                                verify_flags: str = "X509_STRICT",
                                check_hostname: bool = False, server_hostname="",
                                options: str = "ALL,NO_COMPRESSION", ciphers: str = "DEFAULT",
                                server_side: bool = True):
    if server_side and not cert:
        cert = find_ssl_cert(SSL_CERT_FILENAME)
        if not cert:
            raise InitException("you must specify an 'ssl-cert' file to use ssl sockets")
    log = get_ssl_logger()
    log("get_ssl_wrap_socket_context%s", (
        cert, key, ca_certs, ca_data, protocol, client_verify_mode, server_verify_mode, verify_flags,
        check_hostname, server_hostname, options, ciphers, server_side)
        )
    ssl_cert_reqs = parse_ssl_verify_mode(client_verify_mode if server_side else server_verify_mode)
    log(" verify_mode for server_side=%s : %s", server_side, ssl_cert_reqs)
    # parse protocol:
    import ssl
    kwargs: dict[str, bool | str] = {
        "server_side": server_side,
        "do_handshake_on_connect": False,
        "suppress_ragged_eofs": True,
    }
    if not server_side:
        kwargs["server_hostname"] = server_hostname
    proto = parse_ssl_protocol(protocol, server_side)
    context = ssl.SSLContext(proto)
    context.set_ciphers(ciphers)
    if not server_side:
        context.check_hostname = check_hostname
    context.verify_mode = ssl_cert_reqs
    # we can't specify the type hint without depending on the `ssl` module:
    # noinspection PyTypeChecker
    context.verify_flags = parse_ssl_verify_mask(verify_flags)
    context.options = parse_ssl_options_mask(options)
    log(" cert=%s, key=%s", cert, key)
    if cert:
        if cert == "auto":
            # try to locate the cert file from known locations
            cert = find_ssl_cert()
            if not cert:
                raise InitException("failed to automatically locate an SSL certificate to use")
        # Important: keep key_password=None when no password is available
        key_password = key_password or os.environ.get("XPRA_SSL_KEY_PASSWORD")
        log("context.load_cert_chain%s", (cert or None, key or None, key_password))
        try:
            # we must pass a `None` value to ignore `keyfile`:
            context.load_cert_chain(certfile=cert, keyfile=key or None, password=key_password)
        except ssl.SSLError as e:
            log("load_cert_chain", exc_info=True)
            raise InitException(f"SSL error, failed to load certificate chain: {e}") from e
    if ssl_cert_reqs != ssl.CERT_NONE:
        log(" check_hostname=%s, server_hostname=%s", check_hostname, server_hostname)
        purpose = ssl.Purpose.CLIENT_AUTH if server_side else ssl.Purpose.SERVER_AUTH
        if not server_side and context.check_hostname and not server_hostname:
            raise InitException("ssl error: check-hostname is set but server-hostname is not")
        log(" load_default_certs(%s)", purpose)
        context.load_default_certs(purpose)

        # ca-certs:
        if ca_certs == "default":
            ca_certs = ""
        elif ca_certs == "auto":
            ca_certs = find_ssl_cert("ca-cert.pem")
        log(" ca-certs=%s", ca_certs)

        if not ca_certs or ca_certs.lower() == "default":
            log(" using default certs")
            # load_default_certs already calls set_default_verify_paths()
        elif not os.path.exists(ca_certs):
            raise InitException(f"invalid ssl-ca-certs file or directory: {ca_certs}")
        elif os.path.isdir(ca_certs):
            log(" loading ca certs from directory '%s'", ca_certs)
            context.load_verify_locations(capath=ca_certs)
        else:
            log(" loading ca certs from file '%s'", ca_certs)
            if not os.path.isfile(ca_certs):
                raise InitException(f"{ca_certs!r} is not a valid ca file")
            context.load_verify_locations(cafile=ca_certs)
        # ca_data may be hex encoded:
        ca_data = parse_encoded_bin_data(ca_data or "")
        log(" cadata=%s", Ellipsizer(ca_data))
        if ca_data:
            context.load_verify_locations(cadata=ca_data)
    elif check_hostname and not server_side:
        log("cannot check hostname client side with verify mode %s", ssl_cert_reqs)
    return context, kwargs


def do_wrap_socket(tcp_socket, context, **kwargs):
    wrap_socket = context.wrap_socket
    assert tcp_socket
    log = get_ssl_logger()
    log("do_wrap_socket(%s, %s, %s)", tcp_socket, context, kwargs)
    tcp_socket.setblocking(True)
    from ssl import SSLEOFError
    try:
        return wrap_socket(tcp_socket, **kwargs)
    except (InitExit, InitException):
        log.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
        raise
    except SSLEOFError:
        log.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
        return None
    except Exception as e:
        log.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
        raise InitExit(ExitCode.SSL_FAILURE, f"Cannot wrap socket {tcp_socket}: {e}") from None


def ssl_retry(e, ssl_ca_certs: str) -> dict[str, Any]:
    log = get_ssl_logger()
    log("ssl_retry(%s, %s) SSL_RETRY=%s", e, ssl_ca_certs, SSL_RETRY)
    if not SSL_RETRY:
        return {}
    if not isinstance(e, SSLVerifyFailure):
        return {}
    # we may be able to ask the user if he wants to accept this certificate
    verify_code = e.verify_code
    ssl_sock = e.ssl_sock
    msg = str(e)
    del e
    addr = ssl_sock.getpeername()[:2]
    port = addr[-1]
    server_hostname = ssl_sock.server_hostname
    log("ssl_retry: peername=%s, server_hostname=%s", addr, server_hostname)
    if verify_code not in (
            SSL_VERIFY_SELF_SIGNED, SSL_VERIFY_WRONG_HOST,
            SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH,
    ):
        log("ssl_retry: %s not handled here", SSL_VERIFY_CODES.get(verify_code, verify_code))
        return {}
    if not server_hostname:
        log("ssl_retry: no server hostname")
        return {}
    log("ssl_retry: server_hostname=%s, ssl verify_code=%s (%i)",
        server_hostname, SSL_VERIFY_CODES.get(verify_code, verify_code), verify_code)

    def confirm(*args) -> bool:
        from xpra.scripts import pinentry
        ret = pinentry.confirm(*args)
        log("run_pinentry_confirm(..) returned %r", ret)
        return ret

    options = load_ssl_options(server_hostname, port)
    # self-signed cert:
    if verify_code == SSL_VERIFY_SELF_SIGNED:
        if ssl_ca_certs not in ("", "default"):
            log("self-signed cert does not match %r", ssl_ca_certs)
            return {}
        # perhaps we already have the certificate for this hostname
        cert_file = find_ssl_config_file(server_hostname, port, CERT_FILENAME)
        if cert_file:
            log("retrying with %r", cert_file)
            options["ca-certs"] = cert_file
            return options
        # download the certificate data
        import ssl
        try:
            cert_data = ssl.get_server_certificate(addr)
        except ssl.SSLError:
            cert_data = ""
        if not cert_data:
            log.warn("Warning: failed to get server certificate from %s", addr)
            return {}
        log("downloaded ssl cert data for %s: %s", addr, Ellipsizer(cert_data))
        # ask the user if he wants to accept this certificate:
        title = "SSL Certificate Verification Failure"
        prompt = "Do you want to accept this certificate?"
        if not confirm((msg,), title, prompt):
            return {}
        filename = save_ssl_config_file(server_hostname, port,
                                        CERT_FILENAME, "certificate", cert_data.encode("latin1"))
        if not filename:
            log.warn("Warning: failed to save certificate data")
            return {}
        options["ca-certs"] = filename
        save_ssl_options(server_hostname, port, options)
        return options
    if verify_code in (SSL_VERIFY_WRONG_HOST, SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH):
        # ask the user if he wants to skip verifying the host
        title = "SSL Certificate Verification Failure"
        prompt = "Do you want to connect anyway?"
        r = confirm((msg,), title, prompt)
        log("run_pinentry_confirm(..) returned %r", r)
        if r:
            log.info(title)
            log.info(" user chose to connect anyway")
            log.info(" retrying without checking the hostname")
            options["check-hostname"] = False
            save_ssl_options(server_hostname, port, options)
            return options
    return {}



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from typing import Any

from xpra.server.subsystem.stub import StubServerMixin
from xpra.net.common import Packet
from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("network", "ping")

GLib = gi_import("GLib")


class PingServer(StubServerMixin):
    """
    Adds ping handling
    """
    PREFIX = "ping"

    def __init__(self):
        log("ServerBase.__init__()")
        self.pings = False
        self.ping_timer: int = 0

    def init(self, opts) -> None:
        self.pings = opts.pings

    def setup(self) -> None:
        if self.pings > 0:
            self.ping_timer = GLib.timeout_add(1000 * self.pings, self.send_ping)

    def cleanup(self) -> None:
        pt = self.ping_timer
        if pt:
            self.ping_timer = 0
            GLib.source_remove(pt)

    def get_info(self, _proto) -> dict[str, Any]:
        return self.get_caps(None)

    def get_caps(self, _source) -> dict[str, Any]:
        return {
            PingServer.PREFIX: self.pings,
        }

    def get_server_features(self, _source) -> dict[str, Any]:
        return {}

    def send_ping(self) -> bool:
        from xpra.server.source.ping import PingConnection
        for ss in self._server_sources.values():
            if ss.suspended or ss.is_closed():
                continue
            if isinstance(ss, PingConnection):
                ss.ping()
        return True

    def _process_ping_echo(self, proto, packet: Packet) -> None:
        ss = self.get_server_source(proto)
        if ss:
            ss.process_ping_echo(packet)

    def _process_ping(self, proto, packet: Packet) -> None:
        time_to_echo = packet.get_u64(1)
        sid = ""
        if len(packet) >= 4:
            sid = packet.get_str(3)
        ss = self.get_server_source(proto)
        if ss:
            ss.process_ping(time_to_echo, sid)

    def init_packet_handlers(self) -> None:
        self.add_packets(
            "ping", "ping_echo",
        )



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from typing import Any

from xpra.exit_codes import ExitCode
from xpra.net.ssl.common import KEY_FILENAME, CERT_FILENAME, SSL_CERT_FILENAME, SSL_ATTRIBUTES, get_ssl_logger
from xpra.os_util import is_admin, OSX, WIN32, POSIX
from xpra.scripts.config import InitExit
from xpra.util.parsing import TRUE_OPTIONS
from xpra.util.env import osexpand
from xpra.util.io import umask_context, load_binary_file
from xpra.util.str_fn import std, Ellipsizer

KEY_SIZE = 4096
KEY_DAYS = 3650
KEY_SUBJ = "/C=US/ST=Denial/L=Springfield/O=Dis/CN=localhost"


def find_ssl_cert(filename: str = SSL_CERT_FILENAME) -> str:
    log = get_ssl_logger()
    # try to locate the cert file from known locations
    from xpra.platform.paths import get_ssl_cert_dirs  # pylint: disable=import-outside-toplevel
    dirs = get_ssl_cert_dirs()
    log(f"find_ssl_cert({filename}) get_ssl_cert_dirs()={dirs}")
    for d in dirs:
        p = osexpand(d)
        if not os.path.exists(p):
            log(f"ssl cert dir {p!r} does not exist")
            continue
        f = os.path.join(p, filename)
        if not os.path.exists(f):
            log(f"ssl cert {f!r} does not exist")
            continue
        if not os.path.isfile(f):
            log.warn(f"Warning: {f!r} is not a file")
            continue
        if not os.access(f, os.R_OK):
            log.info(f"SSL certificate file {f!r} is not accessible")
            continue
        log(f"found ssl cert {f!r}")
        return os.path.abspath(f)
    return ""


def load_ssl_options(server_hostname: str, port: int) -> dict[str, bool | str]:
    log = get_ssl_logger()
    f = find_ssl_config_file(server_hostname, port, "options")
    options: dict[str, bool | str] = {}
    if f:
        try:
            with open(f, encoding="utf8") as fd:
                for line in fd.readlines():
                    line = line.rstrip("\n\r")
                    if not line or line.startswith("#"):
                        continue
                    parts = line.split("=", 1)
                    if len(parts) != 2:
                        continue
                    k, v = parts
                    if k not in SSL_ATTRIBUTES:
                        log("Warning: unknown SSL attribute %r in %r", k, f)
                        continue
                    # some options use boolean values, convert them back:
                    options[k] = (v.lower() in TRUE_OPTIONS) if k in ("check-hostname",) else v
        except OSError as e:
            log.warn("Warning: failed to read %r: %s", f, e)
    log("load_ssl_options%s=%s (from %r)", (server_hostname, port), options, f)
    return options


def save_ssl_options(server_hostname: str, port, options: dict) -> str:
    log = get_ssl_logger()
    boptions = b"\n".join(("{}={}".format(k.replace("_", "-"), v)).encode("latin1") for k, v in options.items())
    boptions += b"\n"
    f = save_ssl_config_file(server_hostname, port,
                             "options", "configuration options", boptions)
    log("save_ssl_options%s saved to %r", (server_hostname, port, options), f)
    return f


def find_ssl_config_file(hostname: str, port=443, filename=CERT_FILENAME) -> str:
    return do_find_ssl_config_file(hostname, port, filename) or do_find_ssl_config_file(hostname, 0, filename)


def do_find_ssl_config_file(server_hostname: str, port=443, filename=CERT_FILENAME) -> str:
    log = get_ssl_logger()
    from xpra.platform.paths import get_ssl_hosts_config_dirs
    dirs = get_ssl_hosts_config_dirs()
    host_dirname = std(server_hostname, extras="-.:#_")
    if port:
        host_dirname += f"_{port}"
    host_dirs = [os.path.join(osexpand(d), host_dirname) for d in dirs]
    log(f"looking for {filename!r} in {host_dirs}")
    for d in host_dirs:
        f = os.path.join(d, filename)
        if os.path.exists(f):
            log(f"found {f}")
            return os.path.abspath(f)
    return ""


def save_ssl_config_file(server_hostname: str, port=443,
                         filename=CERT_FILENAME, fileinfo="certificate", filedata=b"") -> str:
    log = get_ssl_logger()
    from xpra.platform.paths import get_ssl_hosts_config_dirs
    dirs = get_ssl_hosts_config_dirs()
    host_dirname = std(server_hostname, extras="-.:#_")
    if port:
        host_dirname += f"_{port}"
    host_dirs = [os.path.join(osexpand(d), host_dirname) for d in dirs]
    log(f"save_ssl_config_file%s dirs={dirs}, host_dirname={host_dirname}, host_dirs={host_dirs}",
        (server_hostname, port, filename, fileinfo, Ellipsizer(filedata)), )
    # if there is an existing host config dir, try to use it:
    for d in [x for x in host_dirs if os.path.exists(x)]:
        f = os.path.join(d, filename)
        try:
            with open(f, "wb") as fd:
                fd.write(filedata)
            log.info(f"saved SSL {fileinfo} to {f!r}")
            return f
        except OSError:
            log(f"failed to save SSL {fileinfo} to {f!r}", exc_info=True)
    # try to create a host config dir:
    for d in host_dirs:
        folders = os.path.normpath(d).split(os.sep)
        # we have to be careful and create the 'ssl' dir with 0o700 permissions
        # but any directory above that can use 0o755
        try:
            ssl_dir_index = len(folders) - 1
            while ssl_dir_index > 0 and folders[ssl_dir_index] != "ssl":
                ssl_dir_index -= 1
            if ssl_dir_index > 1:
                folder = folders[0]
                parent = os.path.join(folder, *folders[1:ssl_dir_index - 1])
                ssl_dir = os.path.join(folder, *folders[1:ssl_dir_index])
                os.makedirs(parent, exist_ok=True)
                os.makedirs(ssl_dir, mode=0o700, exist_ok=True)
            os.makedirs(d, mode=0o700)
            f = os.path.join(d, filename)
            with open(f, "wb") as fd:
                fd.write(filedata)
            log.info(f"saved SSL {fileinfo} to {f!r}")
            return f
        except OSError:
            log(f"failed to save cert data to {d!r}", exc_info=True)
    return ""


def get_gen_ssl_cert_dir() -> str:
    if is_admin():
        # running as root, use global location:
        if OSX:
            xpra_dir = "/Library/Application Support/Xpra"
        elif WIN32:
            from xpra.platform.win32.paths import get_program_data_dir
            xpra_dir = os.path.join(get_program_data_dir(), "Xpra")
        else:
            prefix = "" if sys.prefix == "/usr" else sys.prefix
            xpra_dir = f"{prefix}/etc/xpra"
        ssldir = f"{xpra_dir}/ssl"
        if not os.path.exists(xpra_dir):
            os.mkdir(ssldir, 0o777)
        if not os.path.exists(ssldir):
            os.mkdir(ssldir, 0o777)
        return ssldir
    from xpra.platform.paths import get_ssl_cert_dirs
    dirs = [d for d in get_ssl_cert_dirs() if not d.startswith("/etc") and not d.startswith("/usr") and d != "./"]
    # use the first writeable one:
    log = get_ssl_logger()
    log(f"testing ssl dirs: {dirs}")
    for sdir in dirs:
        ssldir = osexpand(sdir)
        if os.path.exists(ssldir) and os.path.isdir(ssldir) and os.access(ssldir, os.W_OK):
            log(f"found writeable ssl dir {ssldir!r}")
            return ssldir
    # we may have to create the parent directories:
    log("no existing ssl dir found, trying to create one")
    for sdir in dirs:
        ssldir = osexpand(sdir)
        head = ssldir
        create_dirs = []
        while head:
            head, tail = os.path.split(head)        #ie: "/home/user/.config/xpra", "ssl"
            log(f"{head=}, {tail=}")
            if tail:
                create_dirs.append(tail)
            else:
                break
            if tail.find("xpra") >= 0:
                # don't create directories above our one
                break
        if create_dirs and head:
            while create_dirs:
                tail = create_dirs.pop()
                head = os.path.join(head, tail)
                try:
                    os.mkdir(head, 0o777)
                except OSError as e:
                    log(f"failed to create {head!r} for {ssldir!r}: {e}")
                    continue
            log(f"created {ssldir!r}")
            return ssldir
    raise RuntimeError("unable to locate or create an ssl certificate directory")


def gen_ssl_cert() -> tuple[str, str]:
    log = get_ssl_logger()
    keypath = find_ssl_cert(KEY_FILENAME)
    certpath = find_ssl_cert(CERT_FILENAME)
    if keypath and certpath:
        log.info("found an existing SSL certificate:")
        log.info(f" {keypath!r}")
        log.info(f" {certpath!r}")
        return keypath, certpath
    from shutil import which
    openssl = which("openssl") or os.environ.get("OPENSSL", "")
    if not openssl:
        raise InitExit(ExitCode.SSL_FAILURE, "cannot find openssl executable")
    ssldir = get_gen_ssl_cert_dir()
    keypath = f"{ssldir}/{KEY_FILENAME}"
    certpath = f"{ssldir}/{CERT_FILENAME}"
    cmd = [
        openssl,
        "req", "-new",
        "-newkey", f"rsa:{KEY_SIZE}",
        "-days", f"{KEY_DAYS}",
        "-nodes", "-x509",
        "-subj", KEY_SUBJ,
        "-keyout", keypath,
        "-out", certpath,
    ]
    openssl_config = ""
    creationflags = 0
    if WIN32:
        from xpra.platform.paths import get_app_dir
        from subprocess import CREATE_NO_WINDOW
        creationflags = CREATE_NO_WINDOW
        openssl_config = os.path.join(get_app_dir(), "etc", "ssl", "openssl.cnf")
    if openssl_config and os.path.exists(openssl_config):
        cmd += ["-config", openssl_config]
    log.info("generating a new SSL certificate:")
    log.info(f" {keypath!r}")
    log.info(f" {certpath!r}")
    log(f"openssl command: {cmd}")
    from subprocess import Popen
    with umask_context(0o022):
        with Popen(cmd, creationflags=creationflags) as p:
            exit_code = p.wait()
    if exit_code != 0:
        raise InitExit(ExitCode.FAILURE, f"openssl command returned {exit_code}")
    key = load_binary_file(keypath)
    cert = load_binary_file(certpath)
    sslcert = key+cert
    sslcertpath = f"{ssldir}/{SSL_CERT_FILENAME}"
    with open(sslcertpath, "wb") as f:
        if POSIX:
            os.fchmod(f.fileno(), 0o600)
        f.write(sslcert)
    return keypath, certpath


def strip_cert(data: bytes) -> bytes:
    BEGIN = b"-----BEGIN CERTIFICATE-----"
    if data.find(BEGIN) >= 0:
        data = BEGIN + data.split(BEGIN, 1)[1]
    END = b"-----END CERTIFICATE-----"
    if data.find(END) > 0:
        data = data.split(END, 1)[0] + END + b"\n"
    return data


def get_ssl_attributes(opts, server_side: bool = True, overrides: dict | None = None) -> dict[str, Any]:
    args: dict[str, Any] = {
        "server-side": server_side,
    }
    for attr in SSL_ATTRIBUTES:
        v = (overrides or {}).get(attr)
        if v is None and opts:
            fn = attr.replace("-", "_")
            ssl_attr = f"ssl_{fn}"  # ie: "ssl_ca_certs"
            v = getattr(opts, ssl_attr)
        args[attr] = v
    return args



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("util")


class SignalEmitter:
    """
    Emulates a subset of GObject signal functions.
    Enough so that subsystems can assume that these functions exist
    by inheriting from this class.
    This allows us to exercise more functionality when running the unit tests.
    """

    def __init__(self):
        self._signal_callbacks: dict[str, list[tuple[Callable, list[Any]]]] = {}

    def connect(self, signal: str, cb: Callable, *args) -> None:
        """ gobject style signal registration """
        log("connect(%s, %s, %s)", signal, cb, args)
        self._signal_callbacks.setdefault(signal, []).append((cb, list(args)))

    def emit(self, signal: str, *extra_args):
        log("%s.emit(%s, %s)", self, signal, extra_args)
        self._fire_callback(signal, extra_args)

    def _fire_callback(self, signal_name: str, extra_args=()) -> None:
        callbacks = self._signal_callbacks.get(signal_name, ())
        log("firing callback for '%s': %s", signal_name, callbacks)
        if not callbacks:
            return
        GLib = gi_import("GLib")
        for cb, args in callbacks:
            with log.trap_error(f"Error processing callback {cb} for {signal_name} packet"):
                all_args = list(args) + list(extra_args)
                GLib.idle_add(cb, self, *all_args)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from xpra.client.base.stub import StubClientMixin
from xpra.log import Logger

log = Logger("event")


class PowerEventClient(StubClientMixin):
    """
    Adds power events callbacks
    """

    def __init__(self):
        self.ui_watcher = None

    def run(self) -> None:
        try:
            self.connect("first-ui-received", self.start_ui_watcher)
        except (TypeError, AttributeError):
            log("no 'first-ui-received' signal")
        from xpra.platform.events import add_handler
        add_handler("suspend", self.suspend)
        add_handler("resume", self.resume)

    def start_ui_watcher(self, _client) -> None:
        if self.ui_watcher:
            return
        from xpra.platform.ui_thread_watcher import get_ui_watcher
        self.ui_watcher = get_ui_watcher()
        assert self.ui_watcher
        self.ui_watcher.start()
        self.ui_watcher.add_resume_callback(self.ui_unpause)
        self.ui_watcher.add_fail_callback(self.ui_pause)

    def cleanup(self) -> None:
        from xpra.platform.events import remove_handler
        remove_handler("suspend", self.suspend)
        remove_handler("resume", self.resume)
        uw = self.ui_watcher
        if uw:
            self.ui_watcher = None
            uw.stop()

    def ui_thread_tick(self):
        uiw = self.ui_watcher
        if uiw:
            uiw.tick()

    def ui_pause(self):
        self.ui_thread_tick()
        self.pause()

    def ui_unpause(self):
        self.ui_thread_tick()
        self.unpause()

    def pause(self) -> None:
        log(f"{self} pause")

    def unpause(self) -> None:
        log(f"{self} unpause")

    def suspend(self) -> None:
        log.info(f"{self} suspending")

    def resume(self) -> None:
        log.info(f"{self} resuming")



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import signal

from xpra.os_util import POSIX
from xpra.log import Logger


def load_pid(pid_file: str) -> int:
    if not pid_file or not os.path.exists(pid_file):
        return 0
    try:
        with open(pid_file, "rb") as f:
            return int(f.read().rstrip(b"\n\r"))
    except (ValueError, OSError) as e:
        log = Logger("util")
        log(f"load_pid({pid_file!r})", exc_info=True)
        log.error(f"Error reading pid from {pid_file!r}: {e}")
        return 0


def write_pid(pidfile: str, pid: int) -> int:
    if pid <= 0:
        raise ValueError(f"invalid pid value {pid}")
    log = Logger("util")
    pidstr = str(pid)
    try:
        with open(pidfile, "w", encoding="latin1") as f:
            if POSIX:
                os.fchmod(f.fileno(), 0o640)
            f.write(f"{pidstr}\n")
            f.flush()
            try:
                fd = f.fileno()
                inode = os.fstat(fd).st_ino
            except OSError as e:
                log("fstat", exc_info=True)
                log.error(f"Error accessing inode of {pidfile!r}: {e}")
                inode = 0
        space = "" if pid == os.getpid() else " "
        log.info(f"{space}wrote pid {pidstr} to {pidfile!r}")
        return inode
    except Exception as e:
        log(f"write_pid({pidfile}, {pid})", exc_info=True)
        log.info(f"Error: failed to write pid {pidstr} to {pidfile!r}")
        log.error(f" {e}")
        return 0


def write_pidfile(pidfile: str) -> int:
    return write_pid(pidfile, os.getpid())


def rm_pidfile(pidfile: str, inode: int) -> bool:
    # verify this is the right file!
    log = Logger("util")
    log("rm_pidfile(%s, %s)", pidfile, inode)
    if inode > 0:
        try:
            i = os.stat(pidfile).st_ino
            log("rm_pidfile: current inode=%i", i)
            if i != inode:
                log.warn(f"Warning: pidfile {pidfile!r} inode has changed")
                log.warn(f" was {inode}, now {i}")
                log.warn(" it would be unsafe to delete it")
                return False
        except OSError as e:
            log("rm_pidfile(%s, %s)", pidfile, inode, exc_info=True)
            log.warn(f"Warning: failed to stat pidfile {pidfile!r}")
            log.warn(f" {e!r}")
            return False
    try:
        os.unlink(pidfile)
        return True
    except OSError as e:
        log("rm_pidfile(%s, %s)", pidfile, inode, exc_info=True)
        log.warn(f"Warning: failed to remove pidfile {pidfile!r}")
        log.warn(f" {e!r}")
        return False


def kill_pid(pid: int, procname: str, sig=signal.SIGTERM) -> None:
    if pid:
        try:
            if pid and pid > 1 and pid != os.getpid():
                os.kill(pid, sig)
        except OSError as e:
            log = Logger("util")
            log.error(f"Error sending {sig!r} signal to {procname!r} pid {pid} {e}")



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.os_util import gi_import
from xpra.scripts.config import InitException
from xpra.net.common import Packet
from xpra.x11.desktop.base import DesktopServerBase
from xpra.server.subsystem.window import WindowsConnection
from xpra.x11.common import X11Event
from xpra.x11.bindings.randr import RandRBindings
from xpra.x11.error import xsync, xlog, XError
from xpra.log import Logger

GObject = gi_import("GObject")
GLib = gi_import("GLib")

log = Logger("server")
metadatalog = Logger("x11", "metadata")
screenlog = Logger("screen")
iconlog = Logger("icon")

MIN_SIZE = 640, 350
MAX_SIZE = 8192, 8192


def get_screen_size() -> tuple[int, int]:
    with xsync:
        return RandRBindings().get_screen_size()


def apply_monitor_config(monitor_defs: dict) -> None:
    try:
        with xsync:
            RandRBindings().set_crtc_config(monitor_defs)
    except XError as e:
        screenlog("apply_monitor_config(%s)", monitor_defs, exc_info=True)
        screenlog.error("Error applying changes to the monitor configuration")
        screenlog.estr(e)


def _adjust_monitor(model, delta_x: int, delta_y: int) -> None:
    screenlog("adjust_monitors(%s, %i, %i)", model, delta_x, delta_y)
    if delta_x == 0 and delta_y == 0:
        return
    x, y = model.get_geometry()[:2]
    new_x = max(0, x + delta_x)
    new_y = max(0, y + delta_y)
    if new_x != x or new_y != y:
        screenlog(f"adjusting monitor {model} from {x},{y} to {new_x},{new_y}")
        mdef = model.get_definition()
        mdef |= {
            "x": new_x,
            "y": new_y,
        }
        model.init(mdef)


class XpraMonitorServer(DesktopServerBase):
    """
        Virtualizes monitors
    """
    __gsignals__ = DesktopServerBase.__common_gsignals__

    def __init__(self):
        super().__init__()
        self.session_type: str = "X11 monitor"
        self.reconfigure_timer: int = 0
        self.reconfigure_locked: bool = False

    def setup(self):
        super().setup()
        with xsync:
            if not RandRBindings().is_dummy16():
                display = os.environ.get("DISPLAY", "")
                raise InitException(f"the vfb display {display!r} cannot virtualize monitors - dummy RandR 1.6 missing")

    def init_randr(self) -> None:
        super().init_randr()
        from xpra.x11.vfb_util import set_initial_resolution
        screenlog(f"init_randr() randr={self.randr}, initial-resolutions={self.initial_resolutions}")
        if not RandRBindings().has_randr() or not self.initial_resolutions:
            return
        res = self.initial_resolutions
        with xlog:
            set_initial_resolution(res, self.dpi or self.default_dpi)

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        if "features" in source.wants:
            capabilities |= {
                "monitor": True,
                "multi-monitors": True,
                "monitors": self.get_monitor_config(),
                "monitors.min-size": MIN_SIZE,
                "monitors.max-size": MAX_SIZE,
            }
        return capabilities

    def configure_best_screen_size(self) -> tuple[int, int]:
        sss = tuple(x for x in self._server_sources.values() if x.ui_client)
        log(f"configure_best_screen_size() sources={sss}")
        if len(sss) != 1:
            screenlog.info(f"screen used by {len(sss)} clients:")
            return get_screen_size()
        ss = sss[0]
        if not getattr(ss, "desktop_fullscreen", False):
            return get_screen_size()
        # try to match this client's layout:
        log("will try to mirror")
        # prevent this monitor layout change
        # from triggering a call to via reconfigure_monitors via reconfigure:
        try:
            self.reconfigure_locked = True
            mdef = self.mirror_client_monitor_layout()
            if mdef:
                self.setting_changed("monitors", mdef)
        except Exception:
            log("Warning: failed to mirror client monitor layout", exc_info=True)
            self.reconfigure_locked = False

        def unlock() -> None:
            self.reconfigure_locked = False

        GLib.timeout_add(1000, unlock)
        return get_screen_size()

    def load_existing_windows(self) -> None:
        with (xlog):
            monitors = RandRBindings().get_monitor_properties()
            screenlog("load_existing_windows() found monitors=%r", monitors)
            outputs = set()
            for i, monitor in monitors.items():
                moutputs = set(monitor.get("outputs", ()))
                # skip this monitor if we already have another one with the same outputs:
                if monitor.get("primary") or not outputs.issuperset(moutputs):
                    outputs.update(moutputs)
                    self.add_monitor_model(i + 1, monitor)
        # does not fire: (because of GTK?)
        # RandR.select_crtc_output_changes()
        # screen = gdk.Screen.get_default()
        # screen.connect("monitors-changed", self.monitors_changed)

    def do_x11_configure_event(self, event: X11Event) -> None:
        # the root window changed,
        # check to see if a monitor has been modified
        # do this via a timer to avoid running multiple times
        # as we get multiple events for the same change
        log("do_x11_configure_event(%s)", event)
        if not self.reconfigure_timer:
            self.reconfigure_timer = GLib.timeout_add(50, self.reconfigure)

    def reconfigure(self) -> None:
        try:
            self.do_reconfigure()
        finally:
            self.reconfigure_timer = 0

    def do_reconfigure(self) -> None:
        # verify that our models are up-to-date,
        # we look for the `crtcs` because that's what tools like `xrandr` can modify easily
        #  ie: `xrandr --output DUMMY1 --mode 1024x768`
        mdefs = {}
        with xlog:
            info = RandRBindings().get_all_screen_properties()
            crtcs = info.get("crtcs")
            outputs = info.get("outputs")
            monitors = info.get("monitors")

            def find_monitor(output_id):
                for minfo in monitors.values():
                    if output_id in minfo.get("outputs", ()):
                        return minfo
                return None

            screenlog("do_reconfigure() crtcs=%s", crtcs)
            screenlog("do_reconfigure() outputs=%s", outputs)
            screenlog("do_reconfigure() monitors=%s", monitors)
            if not crtcs or not outputs:
                return
            for i, crtc_info in crtcs.items():
                # find the monitor for this crtc:
                if crtc_info.get("noutput", 0) != 1:
                    screenlog(f"no outputs on crtc {i}")
                    continue
                output_id = crtc_info.get("outputs")[0]
                output_info = outputs.get(output_id)
                if not output_info:
                    screenlog(f"output {output_id} not found")
                    continue
                if output_info.get("connection") != "Connected":
                    screenlog(f"output {output_id} is not connected")
                    continue
                monitor_info = find_monitor(output_id)
                if not monitor_info:
                    screenlog(f"no monitor found for output id {output_id}")
                    return
                # get the geometry from the crtc:
                mdef: dict[str, int] = {k: v for k, v in crtc_info.items() if k in ("x", "y", "width", "height")}
                # add the millimeter dimensions from the output:
                mdef.update((k, v) for k, v in output_info.items() if k in ("mm-width", "mm-height"))
                # and some monitor attributes:
                mdef.update((k, v) for k, v in output_info.items() if k in ("primary", "automatic", "name"))
                mdefs[i] = mdef
                screenlog(f"do_reconfigure() {i}: {mdef}")
            if self.sync_monitors_to_models(mdefs) and not self.reconfigure_locked:
                self.reconfigure_monitors()
        self.refresh_all_windows()

    def sync_monitors_to_models(self, monitors) -> int:
        # now update the monitor models with this data:
        screenlog("sync_monitors_to_models(%s)", monitors)
        mods = 0
        for i, monitor in monitors.items():
            wid = i + 1
            model = self._id_to_window.get(wid)
            if not model:
                # found a new monitor!
                screenlog("found a new monitor: %s", monitor)
                self.add_monitor_model(wid, monitor)
                continue
            mdef = model.get_definition()
            diff = [k for k in ("x", "y", "width", "height") if monitor.get(k) != mdef.get(k)]
            screenlog("model %i geometry modified %s from %s to %s",
                      wid, diff, [mdef.get(k) for k in diff], [monitor.get(k) for k in diff])
            if diff:
                # re-initialize with new geometry:
                log("was %s, now %s", mdef, monitor)
                model.init(monitor)
                model.emit("resized")
                mods += 1
            if mdef.get("name", "") != monitor.get("name", ""):
                model.name = monitor.get("name")
                screenlog(f"monitor {i} name has changed to {model.name!r}")
                # name is used to generate the window "title":
                model.notify("title")
        return mods

    def add_monitor_model(self, wid: int, monitor):
        from xpra.x11.desktop.monitor_model import MonitorDesktopModel
        model = MonitorDesktopModel(monitor)
        model.setup()
        screenlog("adding monitor model %s", model)
        super().do_add_new_window_common(wid, model)
        model.managed_connect("client-contents-changed", self._contents_changed)
        model.managed_connect("resized", self.monitor_resized)
        model.managed_connect("motion", self._motion_signaled)
        return model

    def monitor_resized(self, model) -> None:
        delta_x, delta_y = model.resize_delta
        wid = self._window_to_id[model]
        w, h = model.get_dimensions()
        screenlog("monitor_resized(%s) size=%s, delta=%s, wid=%#x",
                  model, (w, h), (delta_x, delta_y), wid)
        for ss in self.window_sources():
            ss.resize_window(wid, model, w, h)
        if self.reconfigure_timer:
            # we're in the process of adjusting things
            return
        # we adjust the position of monitors after this one,
        # assuming that they are defined left to right!
        # first the models:
        self._adjust_monitors(wid, delta_x, delta_y)
        self.reconfigure_monitors()

    def reconfigure_monitors(self) -> None:
        # now we can do the virtual crtcs, outputs and monitors
        defs = self.get_monitor_config()
        screenlog("reconfigure_monitors() definitions=%s", defs)
        apply_monitor_config(defs)
        # and tell the client:
        self.setting_changed("monitors", defs)

    def validate_monitors(self) -> None:
        for model in self._id_to_window.values():
            x, y, width, height = model.get_geometry()
            if x + width >= MAX_SIZE[0] or y + height >= MAX_SIZE[1]:
                new_x, new_y = 0, 0
                mdef = model.get_definition()
                mdef |= {
                    "x": new_x,
                    "y": new_y,
                }
                model.init(mdef)

    def _adjust_monitors(self, after_wid: int, delta_x: int, delta_y: int) -> None:
        models = {wid: model for wid, model in self._id_to_window.items() if wid > after_wid}
        screenlog("adjust_monitors(%i, %i, %i) models=%s", after_wid, delta_x, delta_y, models)
        if (delta_x == 0 and delta_y == 0) or not models:
            return
        for wid, model in models.items():
            _adjust_monitor(model, delta_x, delta_y)

    def get_monitor_config(self) -> dict[int, dict]:
        monitor_defs = {}
        for wid, model in self._id_to_window.items():
            monitor = model.get_definition()
            i = wid - 1
            monitor["index"] = i
            monitor_defs[i] = monitor
        return monitor_defs

    def remove_monitor(self, wid: int) -> None:
        model = self._id_to_window.get(wid)
        screenlog("removing monitor for wid %i : %s", wid, model)
        if not model:
            raise ValueError(f"monitor {wid:#x} not found")
        if len(self._id_to_window) <= 1:
            raise RuntimeError("cannot remove the last monitor")
        delta_x = -model.get_definition().get("width", 0)
        delta_y = 0  # model.monitor.get("width", 0)
        model.unmanage()
        wid = self._remove_window(model)
        # adjust the position of the other monitors:
        self._adjust_monitors(wid, delta_x, delta_y)
        self.reconfigure_monitors()

    def add_monitor(self, width: int, height: int) -> None:
        count = len(self._id_to_window)
        if count >= 16:
            raise RuntimeError(f"already too many monitors: {count}")
        if not (isinstance(width, int) and isinstance(height, int)):
            raise ValueError(f"invalid dimension types: {width} ({type(width)}) and {height} ({type(height)})")
        if (width, height) < MIN_SIZE:
            raise ValueError(f"monitor size {width}x{height} is too small, minimum is {MIN_SIZE}")
        if (width, height) > MAX_SIZE:
            raise ValueError(f"monitor size {width}x{height} is too large, maximum is {MIN_SIZE}")

        # find the wid to use:
        # prefer just incrementing the wid, but we cannot go higher than 16

        def rightof(other_wid: int):
            mdef = self._id_to_window[other_wid].get_definition()
            ox = mdef.get("x", 0) + mdef.get("width", 0)
            oy = mdef.get("y", 0)  # +monitor.get("height", 0)
            return ox, oy

        # find a gap we can use in the window ids before 16:
        wid = 1
        for wid in range(1, 16):
            if wid not in self._id_to_window:
                break
        screenlog(f"using {wid=}")
        x = y = 0
        if wid > 1:
            x, y = rightof(wid - 1)
            self._adjust_monitors(wid - 1, width, 0)
        # ensure no monitors end up too far to the right or bottom:
        # (better have them overlap - though we could do something smarter here)
        self.validate_monitors()
        # now we can add our new monitor:
        xdpi = self.xdpi or self.dpi or 96
        ydpi = self.ydpi or self.dpi or 96
        wmm = round(width * 25.4 / xdpi)
        hmm = round(height * 25.4 / ydpi)
        index = wid - 1
        monitor = {
            "index": index,
            "name": f"VFB-{index}",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "mm-width": wmm,
            "mm-height": hmm,
        }
        with xsync:
            model = self.add_monitor_model(wid, monitor)
        self.reconfigure_monitors()
        # send it to the clients:
        for ss in self._server_sources.values():
            if not isinstance(ss, WindowsConnection):
                continue
            self.send_new_desktop_model(model, ss)

    def _process_configure_monitor(self, _proto, packet: Packet) -> None:
        action = packet.get_str(1)
        if action == "remove":
            identifier = packet.get_str(2)
            if identifier == "wid":
                wid = packet.get_wid(3)
            elif identifier == "index":
                # index is zero-based
                wid = packet.get_u32(3) + 1
            else:
                raise ValueError(f"unsupported monitor identifier {identifier!r}")
            self.remove_monitor(wid)
        elif action == "add":
            resolution = packet[2]
            if isinstance(resolution, str):
                from xpra.common import parse_resolution
                resolution = parse_resolution(resolution, self.refresh_rate)
            if not isinstance(resolution, (tuple, list)):
                raise ValueError(f"invalid resolution: {resolution!r} ({type(resolution)}")
            if len(resolution) not in (2, 3) or not all(isinstance(res, int) for res in resolution):
                raise ValueError(f"invalid resolution type: {resolution!r}")
            width, height = resolution[:2]
            self.add_monitor(width, height)
        else:
            raise ValueError(f"unsupported 'configure-monitor' action {action!r}")
        self.refresh_all_windows()

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("configure-monitor", main_thread=True)


GObject.type_register(XpraMonitorServer)



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
from typing import Sequence, Iterable

from xpra.util.str_fn import repr_ellipsized
from xpra.x11.error import xsync


sizeof_long = struct.calcsize(b'@L')


class AlreadyOwned(Exception):
    pass


def xatoms_to_strings(data: bytes) -> Sequence[str]:
    length = len(data)
    if length % sizeof_long != 0:
        raise ValueError(f"invalid length for atom array: {length}, value={repr_ellipsized(data)}")
    natoms = length // sizeof_long
    atoms = struct.unpack(b"@" + b"L" * natoms, data)
    with xsync:
        from xpra.x11.bindings.window import X11WindowBindings
        X11Window = X11WindowBindings()
        return tuple(name for name in (X11Window.get_atom_name(atom) for atom in atoms if atom) if name)


def strings_to_xatoms(data: Iterable[str]) -> bytes:
    with xsync:
        from xpra.x11.bindings.window import X11WindowBindings
        X11Window = X11WindowBindings()
        atom_array = tuple(X11Window.get_xatom(atom) for atom in data if atom)
    return struct.pack(b"@" + b"L" * len(atom_array), *atom_array)


def log_xfixes_error(msg: str) -> None:
    from xpra.util.env import first_time
    from xpra.log import Logger
    log = Logger("x11")
    log("xfixes_error(%s)", msg, exc_info=True)
    if first_time("xfixes"):
        log.warn("Warning: the XFixes extension is not available:")
        log.warn(f" {msg!r}")
        log.warn(" some selection features will be degraded")


def xfixes_selection_input(xid: int, selection: str) -> bool:
    try:
        from xpra.x11.bindings.fixes import XFixesBindings, init_xfixes_events
    except ImportError as e:
        log_xfixes_error(str(e))
        return False
    with xsync:
        xfixes = XFixesBindings()
        if not xfixes.hasXFixes():
            log_xfixes_error("not available")
            return False
        init_xfixes_events()
        xfixes.selectXFSelectionInput(xid, selection)
    return True



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.common import SYNC_ICC
from xpra.util.str_fn import hexstr
from xpra.util.objects import typedict
from xpra.server.subsystem.stub import StubServerMixin
from xpra.log import Logger

log = Logger("screen")


class ICCServer(StubServerMixin):
    PREFIX = "icc"

    def __init__(self):
        StubServerMixin.__init__(self)
        self.icc_profile = b""

    def parse_hello(self, ss, caps, send_ui: bool):
        if send_ui:
            self.set_icc_profile()

    # TODO: should use its own packet rather than getting called by `DisplayManager`:
    def process_icc(self, ss, iccdata: dict[str, Any]):
        if iccdata:
            iccd = typedict(iccdata)
            ss.icc = iccd.get("global", ss.icc)
            ss.display_icc = iccd.get("display", ss.display_icc)
            self.set_icc_profile()

    def last_client_exited(self) -> None:
        self.reset_icc_profile()

    def get_info(self, _proto) -> dict[str, Any]:
        return {"icc": self.get_icc_info()}

    def get_icc_info(self) -> dict[str, Any]:
        icc_info: dict[str, Any] = {
            "sync": SYNC_ICC,
        }
        if SYNC_ICC:
            icc_info["profile"] = hexstr(self.icc_profile)
        return icc_info

    def set_icc_profile(self) -> None:
        if not SYNC_ICC:
            return
        from xpra.x11.xroot_props import root_set
        ui_clients = [s for s in self._server_sources.values() if s.ui_client]
        if len(ui_clients) != 1:
            log("%i UI clients, resetting ICC profile to default", len(ui_clients))
            self.reset_icc_profile()
            return
        icc = typedict(ui_clients[0].icc)
        for x in ("data", "icc-data", "icc-profile"):
            data = icc.bytesget(x)
            if data:
                log("set_icc_profile() icc data for %s: %s (%i bytes)", ui_clients[0], hexstr(data), len(data))
                self.icc_profile = data
                root_set("_ICC_PROFILE", ["u32"], data)
                root_set("_ICC_PROFILE_IN_X_VERSION", "u32", 0 * 100 + 4)  # 0.4 -> 0*100+4*1
                return
        log("no icc data found in %s", icc)
        self.reset_icc_profile()

    def reset_icc_profile(self) -> None:
        log("reset_icc_profile()")
        from xpra.x11.xroot_props import root_del
        root_del("_ICC_PROFILE")
        root_del("_ICC_PROFILE_IN_X_VERSION")
        self.icc_profile = b""



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import errno
import socket
import signal
import threading
from weakref import WeakKeyDictionary
from time import sleep, time, monotonic
from types import FrameType
from typing import Any, NoReturn
from collections.abc import Callable, Sequence, Iterable

from xpra.util.version import XPRA_VERSION, version_str, version_compat_check
from xpra.scripts.server import deadly_signal
from xpra.exit_codes import ExitValue, ExitCode
from xpra.server import ServerExitMode
from xpra.server import features
from xpra.server.auth import AuthenticatedServer
from xpra.util.parsing import TRUE_OPTIONS, FALSE_OPTIONS, parse_bool_or
from xpra.net.common import (
    MAX_PACKET_SIZE, SSL_UPGRADE, PACKET_TYPES,
    is_request_allowed, Packet, has_websocket_handler, HttpResponse, HTTP_UNSUPORTED,
)
from xpra.net.digest import get_caps as get_digest_caps
from xpra.net.socket_util import (
    PEEK_TIMEOUT_MS, SOCKET_PEEK_TIMEOUT_MS,
    setup_local_sockets, add_listen_socket, accept_connection, guess_packet_type,
    peek_connection, close_sockets, SocketListener, socket_fast_read,
)
from xpra.net.bytestreams import (
    Connection, SSLSocketConnection, SocketConnection,
    log_new_connection, pretty_socket, SOCKET_TIMEOUT
)
from xpra.net.net_util import get_network_caps, import_netifaces, get_all_ips
from xpra.net.protocol.factory import get_server_protocol_class
from xpra.net.protocol.socket_handler import SocketProtocol
from xpra.net.protocol.constants import CONNECTION_LOST, GIBBERISH, INVALID
from xpra.net.digest import get_salt, gendigest
from xpra.platform import set_name, threaded_server_init
from xpra.platform.paths import get_app_dir, get_system_conf_dirs, get_user_conf_dirs
from xpra.os_util import (
    force_quit, POSIX,
    get_username_for_uid, get_hex_uuid, getuid, gi_import,
)
from xpra.util.system import register_SIGUSR_signals
from xpra.util.io import load_binary_file, find_libexec_command
from xpra.util.background_worker import add_work_item, quit_worker
from xpra.util.thread import start_thread
from xpra.common import (
    LOG_HELLO, FULL_INFO, DEFAULT_XDG_DATA_DIRS,
    noop, ConnectionMessage, noerr, init_memcheck, subsystem_name,
)
from xpra.util.pysystem import dump_all_frames
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, Ellipsizer, print_nested_dict, nicestr, strtobytes, hexstr
from xpra.util.env import envint, envbool, envfloat, first_time
from xpra.log import Logger

# pylint: disable=import-outside-toplevel

GLib = gi_import("GLib")

log = Logger("server")
netlog = Logger("network")
ssllog = Logger("ssl")
httplog = Logger("http")
wslog = Logger("websocket")

main_thread = threading.current_thread()

MAX_CONCURRENT_CONNECTIONS = envint("XPRA_MAX_CONCURRENT_CONNECTIONS", 100)
SIMULATE_SERVER_HELLO_ERROR = envbool("XPRA_SIMULATE_SERVER_HELLO_ERROR", False)
SERVER_SOCKET_TIMEOUT = envfloat("XPRA_SERVER_SOCKET_TIMEOUT", 0.1)
UNAUTHENTICATED_HELLO_REQUESTS = tuple(
    x.strip() for x in
    os.environ.get("XPRA_UNAUTHENTICATED_HELLO_REQUESTS", "version,connect_test,id").split(",") if x.strip()
)

SYSCONFIG = envbool("XPRA_SYSCONFIG", FULL_INFO > 1)
SHOW_NETWORK_ADDRESSES = envbool("XPRA_SHOW_NETWORK_ADDRESSES", True)
INIT_THREAD_TIMEOUT = envint("XPRA_INIT_THREAD_TIMEOUT", 10)
HTTP_HTTPS_REDIRECT = envbool("XPRA_HTTP_HTTPS_REDIRECT", True)
SSL_PEEK = envbool("XPRA_SSL_PEEK", True)


# class used to distinguish internal errors
# which should not be shown to the client,
# from useful messages we do want to pass on
class ClientException(Exception):
    pass


def force_close_connection(conn) -> None:
    try:
        conn.close()
    except OSError:
        netlog("close_connection()", exc_info=True)


def get_server_base_classes() -> tuple[type, ...]:
    from xpra.server.subsystem.control import ControlHandler
    from xpra.server.glib_server import GLibServer
    from xpra.server.subsystem.platform import PlatformServer
    from xpra.server.subsystem.daemon import DaemonServer
    from xpra.server.subsystem.sessionfiles import SessionFilesServer
    from xpra.server.subsystem.splash import SplashServer
    from xpra.server.subsystem.id import IDServer
    from xpra.server.subsystem.info import InfoServer
    from xpra.server.subsystem.version import VersionServer
    classes: list[type] = [
        GLibServer, PlatformServer, DaemonServer, SessionFilesServer,
        AuthenticatedServer, ControlHandler, SplashServer,
        IDServer, InfoServer, VersionServer,
    ]
    from xpra.server import features
    if features.mdns:
        from xpra.server.subsystem.mdns import MdnsServer
        classes.append(MdnsServer)
    return tuple(classes)


SERVER_BASES = get_server_base_classes()
ServerBaseClass = type("ServerBaseClass", SERVER_BASES, {})
log("ServerBaseClass%s", SERVER_BASES)
SIGNALS: dict[str, int] = {}
for base_class in SERVER_BASES:
    SIGNALS.update(getattr(base_class, "__signals__", {}))
SIGNALS.update({
    "init-thread-ended": 0,
    "running": 0,
})


# noinspection PyMethodMayBeStatic
class ServerCore(ServerBaseClass):
    """
        This is the simplest base class for servers.
        It only handles the connection layer:
        authentication and the initial handshake.
    """
    __signals__ = SIGNALS

    def __init__(self):
        log("ServerCore.__init__()")
        for bc in SERVER_BASES:
            bc.__init__(self)
        self.start_time = time()
        self.hello_request_handlers.update({
            "connect_test": self._handle_hello_request_connect_test,
            "screenshot": self._handle_hello_request_screenshot,
        })
        self.uuid = ""
        self.session_type: str = "unknown"
        self._closing: bool = False
        self._exit_mode = ServerExitMode.UNSET
        # networking bits:
        self.sockets: list[SocketListener] = []
        self._potential_protocols: list[SocketProtocol] = []
        self._rfb_upgrade: int = 0
        self._ssl_attributes: dict = {}
        self._accept_timeout: int = SOCKET_TIMEOUT + 1
        self.ssl_mode: str = ""
        self.ssl_upgrade = False
        self.websocket_upgrade = has_websocket_handler()
        self.ssh_upgrade = False
        self.rdp_upgrade = False
        self.http = False
        self._html: bool = False
        self._www_dir: str = ""
        self._http_headers_dirs: list[str] = []
        self.socket_cleanup: list[Callable] = []
        self.socket_verify_timer: WeakKeyDictionary[SocketProtocol, int] = WeakKeyDictionary()
        self.socket_rfb_upgrade_timer: WeakKeyDictionary[SocketProtocol, int] = WeakKeyDictionary()
        self._max_connections: int = MAX_CONCURRENT_CONNECTIONS
        self._socket_timeout: float = SERVER_SOCKET_TIMEOUT
        self._ws_timeout: int = 5
        self._socket_dir: str = ""
        self._socket_dirs: list = []
        self.unix_socket_paths: list[str] = []
        self.touch_timer: int = 0
        self.session_files: list[str] = [
            "cmdline", "server.env", "config", "server.log*",
            # notifications may use a TMP dir:
            "tmp/*", "tmp",
        ]
        self.session_name = ""

        # Features:
        self.readonly = False
        self.compression_level = 1
        self.exit_with_client = False

        self.init_thread = None

        self._default_packet_handlers: dict[str, Callable] = {}

    def init(self, opts) -> None:
        log("ServerCore.init(%s)", opts)
        self.session_name = str(opts.session_name)
        set_name("Xpra", self.session_name or "Xpra")
        self.unix_socket_paths = []
        self._socket_dir = opts.socket_dir or ""
        if not self._socket_dir and opts.socket_dirs:
            self._socket_dir = opts.socket_dirs[0]
        self._socket_dirs = opts.socket_dirs
        self.compression_level = opts.compression_level
        self.exit_with_client = opts.exit_with_client
        self.readonly = opts.readonly
        self.http = opts.http
        self.websocket_upgrade = opts.websocket_upgrade
        self.ssh_upgrade = opts.ssh_upgrade
        self.rdp_upgrade = opts.rdp_upgrade
        if self.http:
            self.init_html_proxy(opts)
        for bc in SERVER_BASES:
            bc.init(self, opts)
        self.init_ssl(opts)

    def init_ssl(self, opts) -> None:
        self.ssl_mode = opts.ssl
        try:
            from xpra.net.ssl.file import get_ssl_attributes
        except ImportError as e:
            netlog("init_ssl(..) no ssl: %s", e)
            self.ssl_upgrade = False
            return
        self._ssl_attributes = get_ssl_attributes(opts, True)
        netlog("init_ssl(..) ssl attributes=%s", self._ssl_attributes)
        self.ssl_upgrade = opts.ssl.lower() not in FALSE_OPTIONS

    def setup(self) -> None:
        self.init_uuid()
        for bc in SERVER_BASES:
            bc.setup(self)
        self.start_listen_sockets()
        self.init_control_commands()
        self.init_packet_handlers()
        # for things that can take longer:
        self.init_thread = start_thread(target=self.threaded_init, name="server-init-thread")

    def init_control_commands(self) -> None:
        self.add_default_control_commands(features.control)

    ######################################################################
    # run / stop:
    def signal_quit(self, signum, _frame=None) -> None:
        self.closing()
        self.install_signal_handlers(deadly_signal)
        GLib.idle_add(self.clean_quit)
        GLib.idle_add(sys.exit, 128 + signum)

    def clean_quit(self, exit_mode=ServerExitMode.NORMAL) -> None:
        log("clean_quit(%s)", exit_mode)
        if self._exit_mode == ServerExitMode.UNSET:
            self._exit_mode = exit_mode
        GLib.timeout_add(5000, self.force_quit)
        self.closing()
        self.cleanup()
        self.quit_worker()

    def force_quit(self, exit_code: ExitValue = ExitCode.FAILURE) -> NoReturn:
        log("force_quit()")
        force_quit(int(exit_code))

    def quit_worker(self) -> None:
        quit_worker(self.quit)

    def quit(self, exit_mode=ServerExitMode.NORMAL) -> None:
        log("quit(%s)", exit_mode)
        if self._exit_mode == ServerExitMode.UNSET:
            self._exit_mode = exit_mode
        self.closing()
        noerr(sys.stdout.flush)
        stop = self._exit_mode not in (ServerExitMode.EXIT, ServerExitMode.UPGRADE)
        self.late_cleanup(stop=stop)
        self.do_quit()
        log("quit(%s) do_quit done!", exit_mode)
        dump_all_frames()

    def closing(self) -> None:
        if not self._closing:
            self._closing = True
            self.log_closing_message()

    def log_closing_message(self) -> None:
        exiting = self._exit_mode in (ServerExitMode.EXIT, ServerExitMode.UPGRADE)
        log.info("%s server is %s", self.session_type, ["terminating", "exiting"][exiting])

    def install_signal_handlers(self, callback: Callable[[int], None]) -> None:
        def os_signal(signum: signal.Signals | int, _frame: FrameType | None = None) -> None:
            callback(signum)

        signal.signal(signal.SIGINT, os_signal)
        signal.signal(signal.SIGTERM, os_signal)
        register_SIGUSR_signals(GLib.idle_add)

    def threaded_init(self) -> None:
        log("threaded_init() servercore start")
        # platform specific init:
        threaded_server_init()
        init_memcheck()
        add_work_item(self.print_run_info)
        self.threaded_setup()
        self.emit("init-thread-ended")

    def threaded_setup(self) -> None:
        log("threaded_setup() servercore start")
        for c in SERVER_BASES:
            with log.trap_error("Error during threaded setup of %s", c):
                c.threaded_setup(self)
        log("threaded_setup() servercore end")

    def wait_for_threaded_init(self) -> None:
        if not self.init_thread:
            # looks like we didn't make it as far as calling setup()
            log("wait_for_threaded_init() no init thread")
            return
        log("wait_for_threaded_init() %s.is_alive()=%s", self.init_thread, self.init_thread.is_alive())
        if self.init_thread.is_alive():
            log("waiting for initialization thread to complete")
            self.init_thread.join(INIT_THREAD_TIMEOUT)
            if self.init_thread.is_alive():
                log.warn("Warning: initialization thread is still active")

    def get_subsystems(self) -> list[str]:
        return [subsystem_name(c) for c in SERVER_BASES]

    def run(self) -> ExitValue:
        self.install_signal_handlers(self.signal_quit)
        GLib.idle_add(self.server_is_ready)
        self.do_run()
        log("run()")
        return 0

    def server_is_ready(self) -> None:
        self.emit("running")
        log.info("xpra is ready.")
        noerr(sys.stdout.flush)

    def cleanup(self) -> None:
        for bc in SERVER_BASES:
            bc.cleanup(self)
        self.cancel_touch_timer()
        self.cleanup_all_protocols()
        self.do_cleanup()
        self.cleanup_sockets()
        netlog("cleanup() done for server core")

    def do_cleanup(self) -> None:
        # allow just a bit of time for the protocol packet flush
        sleep(0.1)

    def late_cleanup(self, stop=True) -> None:
        for bc in SERVER_BASES:
            bc.late_cleanup(self, stop)
        self.cleanup_all_protocols(force=True)
        self._potential_protocols = []
        from xpra.util.child_reaper import reaper_cleanup
        reaper_cleanup()

    def cleanup_sockets(self) -> None:
        sockets = self.sockets
        self.sockets = []
        close_sockets(sockets)

    def init_uuid(self) -> None:
        # Define a server UUID if needed:
        self.uuid = os.environ.get("XPRA_PROXY_START_UUID", "") or self.get_server_uuid() or get_hex_uuid()
        log(f"server uuid is {self.uuid}")

    def get_server_uuid(self) -> str:
        return ""

    def init_html_proxy(self, opts) -> None:
        httplog(f"init_html_proxy(..) options: html={opts.html!r}")
        # opts.html can contain a boolean, "auto" or the path to the webroot
        www_dir = ""
        html = opts.html or ""
        if html and os.path.isabs(html):
            www_dir = html
            self._html = True
        elif not html or (html.lower() in FALSE_OPTIONS or html.lower() in TRUE_OPTIONS or html.lower() == "auto"):
            self._html = parse_bool_or("html", html)
        else:
            # assume that the html option is a request to open a browser
            self._html = True
            # find a socket we can connect the browser to:
            for mode, bind in {
                "ws": opts.bind_ws,
                "wss": opts.bind_wss,
                "tcp": opts.bind_tcp,
                "ssl": opts.bind_ssl,
            }.items():
                if bind:  # ie: ["0.0.0.0:10000", "127.0.0.1:20000"]
                    from xpra.net.common import open_html_url
                    # open via timeout_add so that the server is running by then,
                    # plus a slight delay so that it can settle down:
                    GLib.timeout_add(1000, open_html_url, html, mode, bind[0])
                    break
            else:
                log.warn("Warning: cannot open html client in a browser")
                log.warn(" no compatible socket found")
        if self._html is not False:  # True or None (for "auto")
            if not (opts.bind_tcp or opts.bind_ws or opts.bind_wss or opts.bind or opts.bind_ssl):
                # we need a socket!
                if self._html:
                    # html was enabled, so log an error:
                    httplog.error("Error: cannot use the html server without a socket")
                self._html = False
        httplog("init_html_proxy(..) html=%s", self._html)
        if not has_websocket_handler():
            if self._html is None:  # auto mode
                httplog.info("html server unavailable, cannot find websocket module")
            elif self._html:
                httplog.error("Error: cannot import websocket connection handler:")
                httplog.error(" the html server will not be available")
            self._html = False
            self.websocket_upgrade = False
        # make sure we have the web root:
        from xpra.platform.paths import get_resources_dir
        if www_dir:
            self._www_dir = str(www_dir)
        else:
            # this is the default value which will be shown in the warning if we don't find a valid one:
            self._www_dir = os.path.abspath(os.path.join(get_resources_dir(), "www"))
            dirs: dict[str, str] = {
                get_resources_dir(): "html5",
                get_resources_dir(): "www",
                get_app_dir(): "www",
            }
            if POSIX:
                xdg_data_dirs = os.environ.get("XDG_DATA_DIRS", DEFAULT_XDG_DATA_DIRS)
                for d in xdg_data_dirs.split(":"):
                    dirs[d] = "www"
                for d in ("/usr/share/xpra", "/usr/local/share/xpra"):
                    dirs[d] = "www"
                dirs["/var/www/xpra"] = "www"
            for ad, d in dirs.items():
                www_dir = os.path.abspath(os.path.join(ad, d))
                if os.path.exists(www_dir):
                    self._www_dir = www_dir
                    httplog("found html5 client in '%s'", www_dir)
                    break
        if not os.path.exists(self._www_dir) and self._html is not False:
            httplog.error("Error: cannot find the html web root")
            httplog.error(f" {self._www_dir!r} does not exist")
            httplog.error(" install the `xpra-html5` package")
            httplog.error(" or turn off the builtin web server using `html=no`")
            self._www_dir = "/usr/share/xpra/www/"
            self._html = False
        if self._html is not False:
            httplog.info(f"serving html content from {self._www_dir!r}")
            self._http_headers_dirs = []
            for d in get_system_conf_dirs():
                self._http_headers_dirs.append(os.path.join(d, "http-headers"))
            if not POSIX or getuid() > 0:
                for d in get_user_conf_dirs():
                    self._http_headers_dirs.append(os.path.join(d, "http-headers"))
            self._http_headers_dirs.append(os.path.abspath(os.path.join(self._www_dir, "../http-headers")))
            self._html = True

    def print_run_info(self) -> None:
        from xpra.common import get_run_info
        for run_info in get_run_info(f"{self.session_type} server"):
            log.info(run_info)

    def notify_new_user(self, ss) -> None:
        pass

    # #####################################################################
    # sockets / connections / packets:
    def init_sockets(self, sockets: list[SocketListener]) -> None:
        self.sockets += sockets

    def init_local_sockets(self, opts, display_name: str, clobber: bool) -> None:
        uid = int(opts.uid)
        gid = int(opts.gid)
        username = get_username_for_uid(uid)
        session_dir = os.environ.get("XPRA_SESSION_DIR", "")
        netlog(f"init_local_sockets(.., {clobber}) {uid=}, {gid=}, {username=!r}, {session_dir=!r}")
        # setup unix domain socket:
        local_sockets = setup_local_sockets(
            opts.bind,  # noqa: F821
            opts.socket_dir, opts.socket_dirs, session_dir,  # noqa: F821
            display_name, clobber,  # noqa: F821
            opts.mmap_group, opts.socket_permissions,  # noqa: F821
            username, uid, gid)
        netlog(f"done setting up local sockets: {local_sockets}")
        self.sockets += local_sockets

        # all unix domain sockets:
        ud_paths = [sock.address for sock in local_sockets if sock.socktype in ("socket", "named-pipe")]
        if ud_paths:
            os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
            if opts.forward_xdg_open:
                xdg_open = find_libexec_command("xdg-open")
                if xdg_open:
                    libexec_dir = os.path.dirname(xdg_open)
                    os.environ["PATH"] = libexec_dir + os.pathsep + os.environ.get("PATH", "")
        else:
            log.warn("Warning: no local server sockets,")
            if opts.forward_xdg_open:
                log.warn(" forward-xdg-open cannot be enabled")
            log.warn(" non-embedded ssh connections will not be available")

    def can_upgrade(self, socktype: str, tosocktype: str, options: dict[str, str]) -> bool:
        to_option_str = options.get(tosocktype, "")
        to_option = to_option_str.lower() in TRUE_OPTIONS
        netlog(f"can_upgrade%s {to_option_str=}, {to_option}", (socktype, tosocktype, options))
        if tosocktype in ("ws", "wss") and not has_websocket_handler():
            return False
        if tosocktype == "rfb":
            # only available with the RFBServer
            return getattr(self, "_rfb_upgrade", False)
        if tosocktype == "rdp":
            return self.rdp_upgrade
        if socktype in ("tcp", "socket", "vsock", "named-pipe"):
            if tosocktype == "ssl":
                if to_option_str:
                    return to_option
                return self.ssl_upgrade
            if tosocktype == "ws":
                if to_option_str:
                    return to_option
                return self.websocket_upgrade
            if tosocktype == "wss":
                if to_option_str:
                    return to_option
                return self.websocket_upgrade and self.ssl_upgrade
        if socktype == "ws" and tosocktype == "wss":
            if to_option_str:
                return to_option
            return self.ssl_upgrade
        if socktype == "ssl" and tosocktype == "wss":
            if to_option_str:
                return to_option
            return self.websocket_upgrade
        if tosocktype == "ssh":
            if to_option_str:
                return to_option
            return self.ssh_upgrade
        return False

    def start_listen_sockets(self) -> None:
        # All right, we're ready to accept customers:
        netlog("start_listen_sockets() sockets=%s", self.sockets)
        for listener in self.sockets:
            netlog(" add_listen_socket: %s", listener)
            GLib.idle_add(self.add_listen_socket, listener)
            address = listener.address
            if listener.socktype == "socket" and address:
                if address.startswith("@"):
                    # abstract sockets can't be 'touch'ed
                    continue
                path = os.path.abspath(address)
                self.unix_socket_paths.append(path)
                netlog("added unix socket path: %s", path)
        if self.unix_socket_paths:
            self.touch_sockets()
            self.touch_timer = GLib.timeout_add(60 * 1000, self.touch_sockets)

    def cancel_touch_timer(self) -> None:
        tt = self.touch_timer
        if tt:
            self.touch_timer = 0
            GLib.source_remove(tt)

    def touch_sockets(self) -> bool:
        netlog("touch_sockets() unix socket paths=%s", self.unix_socket_paths)
        for sockpath in self.unix_socket_paths:
            if sockpath.startswith("@"):
                continue
            if not os.path.exists(sockpath):
                if first_time(f"missing-socket-{sockpath}"):
                    log.warn("Warning: the unix domain socket cannot be found:")
                    log.warn(" '%s'", sockpath)
                    log.warn(" was it deleted by mistake?")
                continue
            try:
                os.utime(sockpath, None)
            except OSError as e:
                if first_time(f"touch-{sockpath}"):
                    netlog.warn(f"Warning: unable to set modified time on {sockpath!r}: {e}")
                netlog("touch_sockets() error on %s", sockpath, exc_info=True)
        return True

    def init_packet_handlers(self) -> None:
        self._default_packet_handlers = {
            "hello": self._process_hello,
            "disconnect": self._process_disconnect,
            "ssl-upgrade": self._process_ssl_upgrade,
            CONNECTION_LOST: self._process_connection_lost,
            GIBBERISH: self._process_gibberish,
            INVALID: self._process_invalid,
        }
        netlog("initializing packet handlers for %s", SERVER_BASES)
        for c in SERVER_BASES:
            c.init_packet_handlers(self)

    def cleanup_all_protocols(self, reason: str | ConnectionMessage = "", force=False) -> None:
        protocols = self.get_all_protocols()
        self.cleanup_protocols(protocols, reason=reason, force=force)

    def get_all_protocols(self) -> Sequence[SocketProtocol]:
        return tuple(self._potential_protocols)

    def cleanup_protocols(self, protocols, reason: str | ConnectionMessage = "", force=False) -> None:
        if not reason:
            reason = {
                ServerExitMode.EXIT: ConnectionMessage.SERVER_EXIT,
                ServerExitMode.UPGRADE: ConnectionMessage.SERVER_UPGRADE,
            }.get(self._exit_mode, ConnectionMessage.SERVER_SHUTDOWN)
        netlog("cleanup_protocols(%s, %s, %s)", protocols, reason, force)
        for protocol in protocols:
            if force:
                self.force_disconnect(protocol)
            else:
                self.disconnect_protocol(protocol, reason)

    def add_listen_socket(self, listener: SocketListener) -> None:
        netlog("add_listen_socket(%s)", listener)
        add_listen_socket(listener, self, self._new_connection)

    def _new_connection(self, listener: SocketListener, handle=0) -> bool:
        """
            Accept the new connection,
            verify that there aren't too many,
            start a thread to dispatch it to the correct handler.
        """
        log("_new_connection%s", (listener, handle))
        if self._closing:
            netlog("ignoring new connection during shutdown")
            return False
        socktype = listener.socktype
        if socktype == "named-pipe":
            from xpra.platform.win32.namedpipes.connection import NamedPipeConnection
            conn: Connection = NamedPipeConnection(listener.socket.pipe_name, handle, listener.options)
            netlog.info("New %s connection received on %s", socktype, conn.target)
            self.make_protocol(socktype, conn, listener.options)
            return True

        conn = accept_connection(listener, self._socket_timeout)
        if conn is None:
            return True
        # limit number of concurrent network connections:
        if socktype != "socket" and len(self._potential_protocols) >= self._max_connections:
            netlog.error("Error: too many connections (%i)", len(self._potential_protocols))
            netlog.error(" ignoring new one: %s", conn.endpoint)
            force_close_connection(conn)
            return True
        # from here on, we run in a thread, so we can poll (peek does)
        start_thread(self.handle_new_connection, f"new-{socktype}-connection", True,
                     args=(listener, conn))
        return True

    def new_conn_err(self, conn, sock, socktype: str, address, packet_type: str, msg="") -> None:
        # not an xpra client
        netlog("new_conn_err", exc_info=True)
        log_fn = netlog.debug if packet_type == "http" else netlog.error
        log_fn("Error: %s connection failed:", socktype)
        if conn.remote:
            log_fn(" packet from %s", pretty_socket(conn.remote))
        if address:
            log_fn(" received on %s", pretty_socket(address))
        if packet_type:
            log_fn(" this packet looks like a '%s' packet", packet_type)
        else:
            log_fn(" invalid packet format, not an xpra client?")
        if msg:
            log_fn(" %s", msg)

        if packet_type == "xpra":
            # try xpra packet format:
            from xpra.net.packet_encoding import pack_one_packet
            packet_data = pack_one_packet(("disconnect", "invalid protocol for this port"))
        elif packet_type == "http":
            # HTTP 400 error:
            packet_data = HTTP_UNSUPORTED
        else:
            packet_data = b"disconnect: connection setup failed"
            if msg:
                packet_data += b", %s?" % strtobytes(msg)
            packet_data += b"\n"
        try:
            # default to plain text:
            sock.settimeout(1)
            conn.write(packet_data)
        except IOError as e:
            netlog("error sending %r: %s", packet_data, e)
        GLib.timeout_add(500, force_close_connection, conn)

    def peek_connection(self, conn) -> bytes:
        timeout = PEEK_TIMEOUT_MS
        if conn.socktype == "rfb":
            # rfb does not send any data, waits for a server packet
            # so don't bother waiting for something that should never come:
            timeout = 0
        elif conn.socktype == "socket":
            timeout = SOCKET_PEEK_TIMEOUT_MS
        peek_data = b""
        if timeout > 0:
            peek_data = peek_connection(conn, timeout)
        return peek_data

    def guess_packet_type(self, peek_data: bytes):
        if peek_data:
            line1 = peek_data.split(b"\n")[0]
            netlog("socket peek hex=%s", hexstr(peek_data[:128]))
            netlog("socket peek line1=%s", Ellipsizer(line1))
        packet_type = guess_packet_type(peek_data)
        netlog("guess_packet_type(..)=%s", packet_type)
        return packet_type

    def handle_new_connection(self, listener: SocketListener, conn) -> None:
        try:
            self.do_handle_new_connection(listener, conn)
        except ValueError as e:
            sock = conn._socket
            self.new_conn_err(conn, sock, conn.socktype, conn.remote, "", str(e))
        except OSError as e:
            netlog("handle_new_connection(%s, %s) socket wrapping failed", listener, conn, exc_info=True)
            sock = conn._socket
            self.new_conn_err(conn, sock, conn.socktype, conn.remote, "", str(e))

    def do_handle_new_connection(self, listener: SocketListener, conn) -> None:
        """
            Use peek to decide what sort of connection this is,
            and start the appropriate handler for it.
        """
        socket_options = listener.options
        sock = conn._socket
        address = conn.remote
        socktype = conn.socktype
        peername = conn.endpoint

        sockname = sock.getsockname()
        target = peername or sockname
        sock.settimeout(self._socket_timeout)

        netlog("handle_new_connection%s sockname=%s, target=%s", (listener, conn), sockname, target)
        # peek so we can detect invalid clients early,
        # or handle non-xpra / wrapped traffic:
        peek_data = self.peek_connection(conn)
        netlog("socket peek=%s", Ellipsizer(peek_data, limit=512))
        packet_type = self.guess_packet_type(peek_data)

        def can_upgrade_to(to_socktype: str) -> bool:
            return self.can_upgrade(socktype, to_socktype, socket_options)

        def conn_err(msg: str) -> tuple[bool, Any, bytes]:
            self.new_conn_err(conn, conn._socket, socktype, address, packet_type, msg)
            return False, None, b""

        if socktype in ("ssl", "wss"):
            # verify that this isn't plain HTTP / xpra:
            if packet_type not in ("ssl", ""):
                conn_err(f"packet is {packet_type!r} and not ssl")
                return
            # always start by wrapping with SSL:
            ssl_conn = self.ssl_wrap(conn, socket_options)
            if not ssl_conn:
                return
            http = socktype == "wss"
            can_peek = SSL_PEEK and (self.ssl_mode.lower() in TRUE_OPTIONS or self.ssl_mode == "auto")
            if can_peek and socktype == "ssl" and can_upgrade_to("wss"):
                # look for HTTPS request to handle:
                line1 = peek_data.split(b"\n")[0]
                if line1.find(b"HTTP/") > 0 or peek_data.find(b"\x08http/") > 0:
                    http = True
                else:
                    ssl_conn.enable_peek()
                    peek_data = peek_connection(ssl_conn)
                    line1 = peek_data.split(b"\n")[0]
                    http = line1.find(b"HTTP/") > 0
                    netlog("looking for 'HTTP' in %r: %s", line1, http)
            if http:
                if not self.http:
                    conn_err("the builtin http server is not enabled")
                    return
                self.start_http_socket(socktype, ssl_conn, socket_options, True, peek_data)
                return
            ssl_conn._socket.settimeout(self._socket_timeout)
            # complete the TLS handshake on the connection-handler thread
            # before handing off to the read thread; with TLS 1.3 the lazy
            # handshake inside recv() can race with post-handshake writes
            # (NewSessionTickets) and cause the read thread to block:
            from xpra.net.ssl.socket import ssl_handshake
            ssl_handshake(ssl_conn._socket)
            log_new_connection(ssl_conn, address)
            self.make_protocol(socktype, ssl_conn, socket_options)
            return

        if socktype == "ws":
            self.handle_ws_socket(conn, socket_options, peek_data)
            return

        if socktype == "rfb":
            self.handle_rfb_connection(conn)
            return

        if socktype == "rdp":
            self.handle_rdp_connection(conn)
            return

        if socktype == "ssh":
            conn = self.handle_ssh_connection(conn, socket_options)
            if not conn:
                return
            peek_data, packet_type = b"", ""

        if socktype in ("tcp", "socket", "named-pipe") and peek_data:
            # see if the packet data is actually xpra or something else
            # that we need to handle via a SSL wrapper or the websocket adapter:
            cont, conn, peek_data = self.may_wrap_socket(conn, socktype, address, socket_options, peek_data)
            netlog("may_wrap_socket(..)=(%s, %s, %r)", cont, conn, Ellipsizer(peek_data))
            if not cont:
                return
            packet_type = guess_packet_type(peek_data)

        pre_read = []
        if socktype == "socket" and not peek_data:
            # try to read from this socket,
            # so short-lived probes don't go through the whole protocol instantiation
            pre = socket_fast_read(conn)
            if not pre:
                netlog("closing %s connection: no data", socktype)
                force_close_connection(conn)
                return
            pre_read.append(pre)
            packet_type = guess_packet_type(pre)

        if packet_type not in ("xpra", ""):
            conn_err("packet type is not xpra")
            return

        # get the new socket object as we may have wrapped it with ssl:
        sock = getattr(conn, "_socket", sock)
        sock.settimeout(self._socket_timeout)
        log_new_connection(conn, address)
        proto = self.make_protocol(socktype, conn, socket_options, pre_read=pre_read)
        if socktype == "tcp" and not peek_data and self._rfb_upgrade > 0:
            t = GLib.timeout_add(self._rfb_upgrade * 1000, self.try_upgrade_to_rfb, proto)
            self.socket_rfb_upgrade_timer[proto] = t

    def ssl_wrap(self, conn, socket_options: dict[str, Any]) -> SSLSocketConnection | None:
        sock = conn._socket
        socktype = conn.socktype
        ssl_sock = self._ssl_wrap_socket(socktype, sock, socket_options)
        ssllog("ssl wrapped socket(%s)=%s", sock, ssl_sock)
        if ssl_sock is None:
            return None
        address = conn.remote
        peername = conn.endpoint
        sockname = ssl_sock.getsockname()
        target = peername or sockname
        ssl_conn = SSLSocketConnection(ssl_sock, sockname, address, target, socktype, socket_options=socket_options)
        ssllog("ssl_wrap(%s, %s)=%s", conn, socket_options, ssl_conn)
        return ssl_conn

    def get_ssl_socket_options(self, socket_options: dict) -> dict[str, Any]:
        ssllog(f"get_ssl_socket_options({socket_options})")
        kwargs = {k.replace("-", "_"): v for k, v in self._ssl_attributes.items()}
        for k, v in socket_options.items():
            # options use '-' but attributes and parameters use '_':
            k = k.replace("-", "_")
            if k.startswith("ssl_"):
                k = k[4:]
                kwargs[k] = v
        ssllog(f"get_ssl_socket_options({socket_options})={kwargs}")
        return kwargs

    def _ssl_wrap_socket(self, socktype: str, sock, socket_options):
        ssllog("ssl_wrap_socket(%s, %s, %s)", socktype, sock, socket_options)
        kwargs = self.get_ssl_socket_options(socket_options)
        try:
            from xpra.net.ssl.socket import ssl_wrap_socket
            ssl_sock = ssl_wrap_socket(sock, **kwargs)
            ssllog("_ssl_wrap_socket(%s, %s)=%s", sock, kwargs, ssl_sock)
            if ssl_sock is None:
                # None means EOF! (we don't want to import ssl bits here)
                ssllog("ignoring SSL EOF error")
            return ssl_sock
        except Exception as e:
            ssllog("SSL error", exc_info=True)
            ssl_paths = [socket_options.get(x, kwargs.get(x)) for x in ("ssl-cert", "ssl-key")]
            cpaths = csv(f"{x!r}" for x in ssl_paths if x)
            log_fn = log.error if first_time(f"ssl-wrap-{socktype}-{socket_options}") else log.debug
            log_fn("Error: failed to create SSL socket")
            log_fn(" from %s socket: %s", socktype, sock)
            if not cpaths:
                log_fn(" no certificate paths specified")
            else:
                log_fn(" check your certificate paths: %s", cpaths)
            log_fn(" %s", e)
            noerr(sock.close)
            return None

    def handle_ws_socket(self, conn, socket_options: dict[str, Any], peek_data: bytes) -> None:
        packet_type = self.guess_packet_type(peek_data)
        if peek_data:
            if packet_type == "ssl" and self.can_upgrade(conn.socktype, "wss", socket_options):
                ssllog("ws socket receiving ssl, upgrading to wss")
                conn = self.ssl_wrap(conn, socket_options)
                if conn is None:
                    return
            elif packet_type not in (None, "http"):
                raise ValueError(f"packet type is {packet_type!r} and not http")
        self.start_http_socket(conn.socktype, conn, socket_options, False, peek_data)

    def handle_ssh_connection(self, conn, socket_options: dict[str, Any]):
        from xpra.server.ssh import make_ssh_server_connection, log as sshlog
        socktype = conn.socktype_wrapped
        none_auth = not self.auth_classes[socktype]
        sshlog("handle_ssh_connection(%s, %s) socktype wrapped=%s", conn, socket_options, socktype)

        def ssh_password_authenticate(username, password) -> bool:
            if not POSIX or getuid() != 0:
                import getpass
                sysusername = getpass.getuser()
                if sysusername != username:
                    sshlog.warn("Warning: ssh password authentication failed,")
                    sshlog.warn(" username does not match:")
                    sshlog.warn(" expected '%s', got '%s'", sysusername, username)
                    return False
            auth_modules = self.make_authenticators(socktype, {"username": username}, conn)
            sshlog("ssh_password_authenticate auth_modules(%s, %s)=%s", username, "*" * len(password), auth_modules)
            for auth in auth_modules:
                # mimic a client challenge:
                digests = ("xor", )
                try:
                    salt, digest = auth.get_challenge(digests)
                    salt_digest = auth.choose_salt_digest(digests)
                    if digest != "xor":
                        raise ValueError(f"unexpected digest {digest}")
                    if salt_digest != "xor":
                        raise ValueError(f"unexpected salt digest {salt_digest}")
                except ValueError as e:
                    sshlog("authentication with %s", auth, exc_info=True)
                    sshlog.warn("Warning: ssh transport cannot use %r authentication:", auth)
                    sshlog.warn(" %s", e)
                    return False
                else:
                    client_salt = get_salt(len(salt))
                    combined_salt = gendigest("xor", client_salt, salt)
                    xored_password = gendigest("xor", password, combined_salt)
                    r = auth.authenticate(xored_password, client_salt)
                    sshlog("%s.authenticate(..)=%s", auth, r)
                    if not r:
                        return False
            return True

        return make_ssh_server_connection(
            conn, socket_options,
            none_auth=none_auth,
            password_auth=ssh_password_authenticate,
            display_name=self.display,
        )

    def try_upgrade_to_rfb(self, proto) -> bool:
        self.cancel_upgrade_to_rfb_timer(proto)
        if proto.is_closed():
            netlog("try_upgrade_to_rfb() protocol is already closed")
            return False
        conn = proto._conn
        netlog("try_upgrade_to_rfb() input_bytecount=%i", conn.input_bytecount)
        if conn.input_bytecount == 0:
            self.upgrade_protocol_to_rfb(proto)
        return False

    def upgrade_protocol_to_rfb(self, proto: SocketProtocol, data: bytes = b"") -> None:
        conn = proto.steal_connection()
        netlog("upgrade_protocol_to_rfb(%s) connection=%s", proto, conn)
        self._potential_protocols.remove(proto)
        proto.wait_for_io_threads_exit(1)
        conn.set_active(True)
        self.handle_rfb_connection(conn, data)

    def cancel_upgrade_to_rfb_timer(self, protocol) -> None:
        t = self.socket_rfb_upgrade_timer.pop(protocol, None)
        if t:
            GLib.source_remove(t)

    def make_protocol(self, socktype: str, conn, socket_options, protocol_class=SocketProtocol, pre_read=()):
        """ create a new xpra Protocol instance and start it """
        log("make_protocol%s", (socktype, conn, socket_options, protocol_class, pre_read))

        def xpra_protocol_class(conn):
            """ adds xpra protocol tweaks after creating the instance """
            protocol = protocol_class(conn, self.process_packet)
            protocol.large_packets.append("info-response")
            return protocol

        return self.do_make_protocol(socktype, conn, socket_options, xpra_protocol_class, pre_read)

    def do_make_protocol(self, socktype: str, conn, socket_options: dict[str, Any], protocol_class,
                         pre_read: Iterable[bytes] = ()) -> SocketProtocol:
        """ create a new Protocol instance and start it """
        netlog("make_protocol%s", (socktype, conn, socket_options, protocol_class, pre_read))
        socktype = socktype.lower()
        protocol = protocol_class(conn)
        protocol._pre_read = list(pre_read)
        protocol.socket_type = socktype
        self._potential_protocols.append(protocol)
        protocol.authenticators = ()
        protocol.invalid_header = self.invalid_header
        # weak dependency on EncryptionServer:
        parse_encryption = getattr(self, "parse_encryption", noop)
        parse_encryption(protocol, socket_options)
        netlog(f"starting {socktype} protocol")
        protocol.start()
        self.schedule_verify_connection_accepted(protocol, self._accept_timeout)
        return protocol

    def may_wrap_socket(self, conn, socktype: str, address, socket_options: dict, peek_data=b"") \
            -> tuple[bool, Any, bytes]:
        """
            Returns:
            * a flag indicating if we should continue processing this connection
            *  (False for webosocket and tcp proxies as they take over the socket)
            * the connection object (which may now be wrapped, ie: for ssl)
            * new peek data (which may now be empty),
        """
        if not peek_data:
            netlog("may_wrap_socket: no data, not wrapping")
            return True, conn, peek_data
        line1 = peek_data.split(b"\n")[0]
        packet_type = self.guess_packet_type(peek_data)
        if packet_type == "xpra":
            netlog("may_wrap_socket: xpra protocol header '%s', not wrapping", peek_data[0])
            # xpra packet header, no need to wrap this connection
            return True, conn, peek_data
        frominfo = pretty_socket(conn.remote)
        netlog("may_wrap_socket(..) peek_data=%s from %s", Ellipsizer(peek_data), frominfo)
        netlog("may_wrap_socket(..) peek_data got %i bytes", len(peek_data))
        netlog("may_wrap_socket(..) packet_type=%s", packet_type)

        def can_upgrade_to(to_socktype: str) -> bool:
            return self.can_upgrade(socktype, to_socktype, socket_options)

        def conn_err(msg: str) -> tuple[bool, Any, bytes]:
            self.new_conn_err(conn, conn._socket, socktype, address, packet_type, msg)
            return False, None, b""

        if packet_type == "ssh":
            if not can_upgrade_to("ssh"):
                conn_err("ssh upgrades are not enabled")
                return False, None, b""
            conn = self.handle_ssh_connection(conn, socket_options)
            return conn is not None, conn, b""
        if packet_type == "ssl":
            sock, sockname, address, endpoint = conn._socket, conn.local, conn.remote, conn.endpoint
            sock = self._ssl_wrap_socket(socktype, sock, socket_options)
            if sock is None:
                return False, None, b""
            conn = SSLSocketConnection(sock, sockname, address, endpoint, "ssl", socket_options=socket_options)
            conn.socktype_wrapped = socktype
            # we cannot peek on SSL sockets, just clear the unencrypted data:
            http = False
            ssl_mode = (socket_options.get("ssl-mode", "") or self.ssl_mode).lower()
            if ssl_mode == "tcp":
                http = False
            elif ssl_mode == "www":
                http = True
            elif ssl_mode == "auto" or ssl_mode in TRUE_OPTIONS and can_upgrade_to("wss"):
                # use the header to guess:
                if line1.find(b"HTTP/") > 0 or peek_data.find(b"\x08http/1.1") > 0:
                    http = True
                else:
                    conn.enable_peek()
                    peek_data = peek_connection(conn)
                    line1 = peek_data.split(b"\n")[0]
                    http = line1.find(b"HTTP/") > 0
            ssllog("may_wrap_socket SSL: %s, ssl mode=%s, http=%s", conn, ssl_mode, http)
            to_socktype = "wss" if http else "ssl"
            if not can_upgrade_to(to_socktype):
                conn_err(f"{to_socktype} upgrades are not enabled for this socket")
                return False, None, b""
            is_ssl = True
        else:
            http = line1.find(b"HTTP/") > 0
            is_ssl = False
        if http:
            if not has_websocket_handler():
                conn_err("websocket module is not installed")
                return False, None, b""
            ws_protocol = "wss" if is_ssl else "ws"
            if not can_upgrade_to(ws_protocol):
                conn_err(f"{ws_protocol} upgrades are not enabled for this socket")
                return False, None, b""
            self.start_http_socket(socktype, conn, socket_options, is_ssl, peek_data)
            return False, conn, b""
        return True, conn, peek_data

    def invalid_header(self, proto: SocketProtocol, data: bytes, msg="") -> None:
        netlog("invalid header: %s, input_packetcount=%s, websocket_upgrade=%s, ssl=%s",
               Ellipsizer(data), proto.input_packetcount, self.websocket_upgrade, bool(self._ssl_attributes))
        if data == b"RFB " and self._rfb_upgrade > 0:
            netlog("RFB header, trying to upgrade protocol")
            self.cancel_upgrade_to_rfb_timer(proto)
            self.upgrade_protocol_to_rfb(proto, data)
            return
        packet_type = self.guess_packet_type(data)
        # RFBServerProtocol doesn't support `steal_connection`:
        if packet_type == "http" and hasattr(proto, "steal_connection"):
            # try again to wrap this socket:
            bufs = [data]

            def addbuf(buf) -> None:
                bufs.append(buf)

            conn = proto.steal_connection(addbuf)
            self.cancel_verify_connection_accepted(proto)
            self.cancel_upgrade_to_rfb_timer(proto)
            netlog(f"stole connection: {type(conn)}")
            # verify that it is not wrapped yet:
            if isinstance(conn, SocketConnection) and conn.socktype_wrapped == conn.socktype:
                conn.enable_peek(b"".join(bufs))
                conn.set_active(True)
                cont, conn, peek_data = self.may_wrap_socket(conn, conn.socktype, conn.info, conn.options,
                                                             b"".join(bufs))
                netlog("wrap : may_wrap_socket(..)=(%s, %s, %r)", cont, conn, Ellipsizer(peek_data))
                if not cont:
                    return
            if conn:
                # the connection object is now removed from the protocol object,
                # so we have to close it explicitly if we have not wrapped it successfully:
                force_close_connection(conn)
        proto._invalid_header(proto, data, msg)

    # #####################################################################
    # http / websockets:
    def start_http_socket(self, socktype: str, conn, socket_options: dict[str, Any], is_ssl: bool = False,
                          peek_data: bytes = b""):
        frominfo = pretty_socket(conn.remote)
        line1 = peek_data.split(b"\n")[0]
        http_proto = "http" + ["", "s"][int(is_ssl)]
        netlog("start_http_socket(%s, %s, %s, %s, ..) http proto=%s, line1=%r",
               socktype, conn, socket_options, is_ssl, http_proto, line1)
        if line1.startswith(b"GET ") or line1.startswith(b"POST "):
            parts = line1.decode("latin1").split(" ")
            httplog("New %s %s request received from %s for '%s'", http_proto, parts[0], frominfo, parts[1])
            tname = parts[0] + "-request"
            req_info = http_proto + " " + parts[0]
        else:
            httplog("New %s connection received from %s", http_proto, frominfo)
            req_info = "wss" if is_ssl else "ws"
            tname = f"{req_info}-proxy"
        # we start a new thread,
        # only so that the websocket handler thread is named correctly:
        start_thread(self.start_http, f"{tname}-for-{frominfo}",
                     daemon=True, args=(socktype, conn, socket_options, is_ssl, req_info, line1, conn.remote))

    def start_http(self, socktype: str, conn, socket_options: dict[str, Any], is_ssl: bool, req_info: str,
                   line1: bytes, frominfo) -> None:
        httplog("start_http(%s, %s, %s, %s, %s, %r, %s) www dir=%s, headers dir=%s",
                socktype, conn, socket_options, is_ssl, req_info, line1, frominfo,
                self._www_dir, self._http_headers_dirs)
        try:
            from xpra.net.websockets.handler import WebSocketRequestHandler
            sock = conn._socket
            sock.settimeout(self._ws_timeout)

            def new_websocket_client(wsh) -> None:
                from xpra.net.websockets.protocol import WebSocketProtocol
                wslog("new_websocket_client(%s) socket=%s, headers=%s", wsh, sock, wsh.headers)
                newsocktype = "wss" if is_ssl else "ws"
                socket_options["http-headers"] = dict(wsh.headers)
                self.make_protocol(newsocktype, conn, socket_options, WebSocketProtocol)

            scripts = self.get_http_scripts()
            conn.socktype = "wss" if is_ssl else "ws"
            redirect_https = False
            if HTTP_HTTPS_REDIRECT and req_info not in ("ws", "wss"):
                ssl_mode = (socket_options.get("ssl-mode", "") or self.ssl_mode).lower()
                redirect_https = not is_ssl and ssl_mode in TRUE_OPTIONS
            WebSocketRequestHandler(sock, frominfo, new_websocket_client,
                                    self._www_dir, self._http_headers_dirs, scripts,
                                    redirect_https)
            return
        except (OSError, ValueError) as e:
            # don't log a full backtrace for `SSLV3_ALERT_CERTIFICATE_UNKNOWN`
            # and detect it without importing the ssl module
            exc_info = True
            if type(e).__name__ == "SSLError":
                reason = getattr(e, "reason", "")
                log(f"ssl socket error: {e}, library=%s, reason=%s", getattr(e, "library", ""), reason)
                if reason == "SSLV3_ALERT_CERTIFICATE_UNKNOWN":
                    exc_info = False
            httplog("start_http%s", (socktype, conn, is_ssl, req_info, frominfo), exc_info=exc_info)
            err = e.args[0]
            if err == 1 and line1 and line1[0] == 0x16:
                log_fn = httplog.debug
            elif err in (errno.EPIPE, errno.ECONNRESET):
                log_fn = httplog.debug
            else:
                log_fn = httplog.error
                log_fn("Error: %s request failure", req_info)
                log_fn(" errno=%s", err)
            log_fn(" for client %s:", pretty_socket(frominfo))
            if line1 and line1[0] >= 128 or line1[0] == 0x16:
                log_fn(" request as hex: '%s'", hexstr(line1))
            else:
                log_fn(" request: %r", line1)
            log_fn(" %s", e)
        force_close_connection(conn)

    def get_http_scripts(self) -> dict[str, Callable[[str], HttpResponse]]:
        # loose coupling with xpra.server.subsystem.http:
        return getattr(self, "_http_scripts", {})

    def is_timedout(self, protocol: SocketProtocol) -> bool:
        # subclasses may override this method (ServerBase does)
        v = not protocol.is_closed() and protocol in self._potential_protocols
        netlog("is_timedout(%s)=%s", protocol, v)
        return v

    def schedule_verify_connection_accepted(self, protocol: SocketProtocol, timeout: int = 60) -> None:
        t = GLib.timeout_add(timeout * 1000, self.verify_connection_accepted, protocol)
        self.socket_verify_timer[protocol] = t

    def verify_connection_accepted(self, protocol: SocketProtocol) -> None:
        self.cancel_verify_connection_accepted(protocol)
        if self.is_timedout(protocol):
            conn = getattr(protocol, "_conn", None)
            elapsed = round(monotonic() - protocol.start_time)
            messages = [
                "Error: connection timed out",
                " " + str(conn or protocol),
                f" after {elapsed} seconds",
            ]
            if conn:
                messages += [
                    f" sent {conn.output_bytecount} bytes",
                    f" received {conn.input_bytecount} bytes",
                ]
                if conn.input_bytecount == 0:
                    try:
                        data = conn.peek(200)
                    except (OSError, ValueError):
                        log("cannot peek on %s", conn)
                        data = b""
                    if data:
                        messages.append(f" read buffer={data!r}")
                        packet_type = self.guess_packet_type(data)
                        if packet_type:
                            messages.append(f" looks like {packet_type!r}")
                    else:
                        # no data was ever received,
                        # this can happen with probes or browser connections,
                        # log at debug level only to avoid spamming the log:
                        for msg in messages:
                            netlog(msg)
                        messages = []
            for msg in messages:
                netlog.error(msg)
            self.send_disconnect(protocol, ConnectionMessage.LOGIN_TIMEOUT)

    def cancel_verify_connection_accepted(self, protocol: SocketProtocol) -> None:
        t = self.socket_verify_timer.pop(protocol, None)
        if t:
            GLib.source_remove(t)

    def send_disconnect(self, proto: SocketProtocol, *reasons) -> None:
        netlog("send_disconnect(%s, %s)", proto, reasons)
        self.cancel_verify_connection_accepted(proto)
        self.cancel_upgrade_to_rfb_timer(proto)
        if proto.is_closed():
            return
        proto.send_disconnect(reasons)
        GLib.timeout_add(1000, self.force_disconnect, proto)

    def force_disconnect(self, proto: SocketProtocol) -> None:
        netlog("force_disconnect(%s)", proto)
        self.cleanup_protocol(proto)
        self.cancel_verify_connection_accepted(proto)
        self.cancel_upgrade_to_rfb_timer(proto)
        proto.close()

    def disconnect_client(self, protocol: SocketProtocol, reason: str | ConnectionMessage, *extra) -> None:
        netlog("disconnect_client(%s, %s, %s)", protocol, reason, extra)
        if protocol and not protocol.is_closed():
            self.disconnect_protocol(protocol, reason, *extra)

    def disconnect_protocol(self, protocol: SocketProtocol, *reasons) -> None:
        netlog("disconnect_protocol(%s, %s)", protocol, reasons)
        i = nicestr(reasons[0])
        if len(reasons) > 1:
            i += " (%s)" % csv(reasons[1:])
        proto_info = f" {protocol}"
        try:
            conn = protocol._conn
            info = conn.get_info()
            endpoint = info.get("endpoint")
            if endpoint:
                proto_info = " " + pretty_socket(endpoint)
            else:
                proto_info = " " + pretty_socket(conn.local)
        except (KeyError, AttributeError):
            pass
        self.cancel_verify_connection_accepted(protocol)
        self.cancel_upgrade_to_rfb_timer(protocol)
        if not protocol.is_closed():
            self._log_disconnect(protocol, "Disconnecting client%s:", proto_info)
            self._log_disconnect(protocol, " %s", i)
            protocol.send_disconnect(reasons)
        self.cleanup_protocol(protocol)

    def cleanup_protocol(self, protocol: SocketProtocol) -> None:
        """ some subclasses perform extra cleanup here """

    def _process_disconnect(self, proto: SocketProtocol, packet: Packet) -> None:
        info = packet.get_str(1)
        if len(packet) > 2:
            info += " (%s)" % csv(str(x) for x in packet[2:])
        # only log protocol info if there is more than one client:
        proto_info = self._disconnect_proto_info(proto)
        self._log_disconnect(proto, "client%s has requested disconnection: %s", proto_info, info)
        self.disconnect_protocol(proto, ConnectionMessage.CLIENT_REQUEST)

    def _log_disconnect(self, _proto: SocketProtocol, *args) -> None:
        netlog.info(*args)

    def _disconnect_proto_info(self, _proto) -> str:
        # overridden in server_base in case there is more than one protocol
        return ""

    def _process_connection_lost(self, proto: SocketProtocol, packet: Packet) -> None:
        netlog("process_connection_lost(%s, %s)", proto, packet)
        self.cancel_verify_connection_accepted(proto)
        self.cancel_upgrade_to_rfb_timer(proto)
        if proto in self._potential_protocols:
            if not proto.is_closed():
                self._log_disconnect(proto, "Connection lost")
                for extra in packet[1:]:
                    self._log_disconnect(proto, " %s", extra)
            self._potential_protocols.remove(proto)
        self.cleanup_protocol(proto)

    def _process_gibberish(self, proto: SocketProtocol, packet: Packet) -> None:
        message = packet.get_str(1)
        data = packet.get_bytes(2)
        netlog("Received uninterpretable nonsense from %s: %s", proto, message)
        netlog(" data: %s", Ellipsizer(data))
        self.disconnect_client(proto, message)

    def _process_invalid(self, protocol: SocketProtocol, packet: Packet) -> None:
        message = packet.get_str(1)
        data = packet.get_bytes(2)
        netlog(f"Received invalid packet: {message}")
        netlog(" data: %s", Ellipsizer(data))
        self.disconnect_client(protocol, message)

    # #####################################################################
    # hello / authentication:
    def send_version_info(self, proto: SocketProtocol, full: bool = False) -> None:
        version = version_str() if (full and FULL_INFO) else XPRA_VERSION.split(".", 1)[0]
        proto.send_now(Packet("hello", {"version": version}))
        # client is meant to close the connection itself, but just in case:
        GLib.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.DONE, "version sent")

    def _process_hello(self, proto: SocketProtocol, packet: Packet) -> None:
        capabilities = packet.get_dict(1)
        c = typedict(capabilities)
        if LOG_HELLO:
            netlog.info(f"hello from {proto}:")
            print_nested_dict(c, print_fn=netlog.info)
        proto.set_compression_level(c.intget("compression_level", self.compression_level))
        proto.enable_compressor_from_caps(c)
        if not proto.enable_encoder_from_caps(c):
            # this should never happen:
            # if we got here, we parsed a packet from the client!
            # (maybe the client used an encoding it claims not to support?)
            self.disconnect_client(proto, ConnectionMessage.PROTOCOL_ERROR, "failed to negotiate a packet encoder")
            return

        log("process_hello: capabilities=%s", capabilities)
        request = c.strget("request")
        if request in UNAUTHENTICATED_HELLO_REQUESTS and self.do_handle_hello_request(request, proto, c):
            return

        # verify version:
        remote_version = c.strget("version")
        verr = version_compat_check(remote_version)
        if verr is not None:
            self.disconnect_client(proto, ConnectionMessage.VERSION_ERROR, f"incompatible version: {verr!r}")
            proto.close()
            return

        # try to auto upgrade to ssl:
        packet_types = c.strtupleget("packet-types", ())
        encryption_caps = c.dictget("encryption")
        conn = getattr(proto, "_conn", None)
        if SSL_UPGRADE and not encryption_caps and "ssl-upgrade" in packet_types and conn and conn.socktype in ("tcp",):
            options = conn.options
            if options.get("ssl-upgrade", "yes").lower() in TRUE_OPTIONS:
                ssl_options = self.get_ssl_socket_options(options)
                cert = ssl_options.get("cert", "")
                if cert:
                    log.info(f"sending ssl upgrade for {conn}")
                    cert_data = load_binary_file(cert)
                    ssl_attrs = {"cert-data": cert_data}
                    proto.send_now(Packet("ssl-upgrade", ssl_attrs))
                    return

        # if we're here, then we should have read everything that was sent by the client,
        # including any data we potentially peeked at:
        disable_peek = getattr(conn, "disable_peek", noop)
        disable_peek()

        # this will call auth_verified if successful
        # it may also just send challenge packets,
        # in which case we'll end up here parsing the hello again
        start_thread(AuthenticatedServer.verify_auth, "authenticate connection", daemon=True, args=(self, proto, packet, c))

    def _process_ssl_upgrade(self, proto: SocketProtocol, packet: Packet) -> None:
        socktype = proto._conn.socktype
        new_socktype = {"tcp": "ssl", "ws": "wss"}.get(socktype)
        if not new_socktype:
            raise ValueError(f"cannot upgrade {socktype} to ssl")
        self.cancel_verify_connection_accepted(proto)
        self.cancel_upgrade_to_rfb_timer(proto)
        if proto in self._potential_protocols:
            self._potential_protocols.remove(proto)
        ssllog("ssl-upgrade: %s", packet[1:])
        conn = proto.steal_connection()
        # threads should be able to terminate immediately
        # as there's no traffic yet:
        ioe = proto.wait_for_io_threads_exit(1)
        if not ioe:
            self.disconnect_protocol(proto, "failed to terminate network threads for ssl upgrade")
            force_close_connection(conn)
            return
        options = conn.options
        socktype = conn.socktype
        ssl_sock = self._ssl_wrap_socket(socktype, conn._socket, options)
        if not ssl_sock:
            self.disconnect_protocol(proto, "failed to upgrade socket to ssl")
            force_close_connection(conn)
            return
        ssl_conn = SSLSocketConnection(ssl_sock, conn.local, conn.remote, conn.endpoint, "ssl", socket_options=options)
        ssl_conn.socktype_wrapped = socktype
        protocol_class = get_server_protocol_class(new_socktype)
        self.make_protocol(new_socktype, ssl_conn, options, protocol_class)
        ssllog.info("upgraded %s to %s", conn, new_socktype)

    def call_hello_oked(self, proto: SocketProtocol, c: typedict, auth_caps: dict) -> None:
        try:
            if SIMULATE_SERVER_HELLO_ERROR:
                raise RuntimeError("Simulating a server error")
            self.hello_oked(proto, c, auth_caps)
        except ClientException as e:
            log("call_hello_oked(%s, %s, %s)", proto, Ellipsizer(c), auth_caps, exc_info=True)
            log.error("Error setting up new connection for")
            log.error(" %s:", proto)
            log.estr(e)
            self.disconnect_client(proto, ConnectionMessage.CONNECTION_ERROR, str(e))
        except Exception as e:
            # log exception but don't disclose internal details to the client
            log.error("server error processing new connection from %s: %s", proto, e, exc_info=True)
            self.disconnect_client(proto, ConnectionMessage.CONNECTION_ERROR, "error accepting new connection")

    def hello_oked(self, proto: SocketProtocol, c: typedict, _auth_caps: dict) -> bool:
        if self._closing:
            self.disconnect_client(proto, ConnectionMessage.SERVER_EXIT, "server is shutting down")
            return True
        request = c.strget("request")
        if request and self.handle_hello_request(request, proto, c):
            return True
        return False

    def handle_hello_request(self, request: str, proto, caps: typedict) -> bool:
        if not is_request_allowed(proto, request):
            msg = f"{request!r} requests are not enabled for this connection"
            log.error(f"Error: {msg}")
            self.send_disconnect(proto, ConnectionMessage.PERMISSION_ERROR, msg)
            return True
        return self.do_handle_hello_request(request, proto, caps)

    def do_handle_hello_request(self, request: str, proto, caps: typedict) -> bool:
        handler = self.hello_request_handlers.get(request)
        log("do_handle_hello_request(%s, %s, ..) handler=%s", request, proto, handler)
        if not handler:
            log.error(f"Error: no handler for hello request {request!r}")
            return False
        return handler(proto, caps)

    def _handle_hello_request_connect_test(self, proto, caps: typedict) -> bool:
        ctr = caps.strget("connect_test_request")
        response = {"connect_test_response": ctr}
        proto.send_now(Packet("hello", response))
        return True

    def _handle_hello_request_screenshot(self, proto, _caps: typedict) -> bool:
        packet = self.make_screenshot_packet()
        proto.send_now(packet)
        return True

    def accept_client(self, proto: SocketProtocol, c: typedict) -> None:
        # max packet size from client (the biggest we can get are clipboard packets)
        netlog("accept_client(%s, %s)", proto, c)
        # note: when uploading files, we send them in chunks smaller than this size
        proto.max_packet_size = MAX_PACKET_SIZE
        proto.parse_remote_caps(c)
        self.accept_protocol(proto)

    def accept_protocol(self, proto: SocketProtocol) -> None:
        if proto in self._potential_protocols:
            self._potential_protocols.remove(proto)
        self.cancel_verify_connection_accepted(proto)
        self.cancel_upgrade_to_rfb_timer(proto)

    def make_hello(self, source) -> dict[str, Any]:
        now = time()
        capabilities = get_network_caps(FULL_INFO)
        for bc in SERVER_BASES:
            capabilities |= bc.get_caps(self, source)
        capabilities |= get_digest_caps()
        capabilities |= {
            "start_time": int(self.start_time),
            "current_time": int(now),
            "elapsed_time": int(now - self.start_time),
            "server_type": "core",
            "server.mode": self.session_type,
        }
        if FULL_INFO > 0:
            capabilities["hostname"] = socket.gethostname()
        if source is None or "features" in source.wants:
            capabilities |= {
                "readonly-server": True,
                "readonly": self.readonly,
            }
            server_log = os.environ.get("XPRA_SERVER_LOG", "")
            if server_log:
                capabilities["server-log"] = server_log
        if source and "packet-types" in source.wants:
            capabilities["packet-types"] = PACKET_TYPES
        if self.session_name:
            capabilities["session_name"] = self.session_name
        return capabilities

    def get_threaded_info(self, proto, **kwargs) -> dict[str, Any]:
        log("ServerCore.get_threaded_info(%s, %s)", proto, kwargs)
        start = monotonic()
        # this function is for non UI thread info, see also: `get_ui_info`
        info = {}
        subsystems = kwargs.get("subsystems", ())
        for bc in SERVER_BASES:
            if subsystems and subsystem_name(bc) not in subsystems:
                continue
            info.update(bc.get_info(self, proto))

        # soft dependency on Info subsystem:
        if hasattr(self, "get_server_info"):
            authenticated = bool(proto and proto.authenticators)
            socktype = getattr(getattr(proto, "_conn", None), "socktype", "")
            local = socktype in ("socket", "named-pipe")
            log("get_threaded_info(%s, %s) authenticated=%s, subsystems=%s, local=%s",
                proto, kwargs, authenticated, subsystems, local)
            full = FULL_INFO > 0 or authenticated or local
            info = self.get_server_info(full)
            if full and (not subsystems or "threads" in subsystems):
                from xpra.server.subsystem.info import get_thread_info
                info["threads"] = get_thread_info(proto)

        end = monotonic()
        log("ServerCore.get_info took %ims", (end - start) * 1000)
        return info

    def get_socket_info(self) -> dict[str, Any]:
        si: dict[str, Any] = {}

        def add_listener(socktype: str, info) -> None:
            si.setdefault(socktype, {}).setdefault("listeners", []).append(info)

        def add_address(socktype: str, address, port: int) -> None:
            addresses = si.setdefault(socktype, {}).setdefault("addresses", [])
            if (address, port) not in addresses:
                addresses.append((address, port))
            if socktype == "tcp":
                if self.websocket_upgrade:
                    add_address("ws", address, port)
                if self._ssl_attributes:
                    add_address("ssl", address, port)
                if self.ssh_upgrade:
                    add_address("ssh", address, port)
            if socktype == "ws" and self._ssl_attributes:
                add_address("wss", address, port)

        netifaces = import_netifaces()
        for sock in self.sockets:
            socktype = sock.socktype
            address = sock.address
            if not address:
                continue
            add_listener(socktype, address)
            if not SHOW_NETWORK_ADDRESSES:
                continue
            if socktype not in ("tcp", "ssl", "ws", "wss", "ssh"):
                # we expose addresses only for TCP sockets
                continue
            upnp_address = sock.options.get("upnp-address", ())
            if upnp_address:
                add_address(socktype, *upnp_address)
            if len(address) != 2 or not isinstance(address[0], str) or not isinstance(address[1], int):
                # unsupported listener info format
                continue
            host, port = address
            if host not in ("0.0.0.0", "::/0", "::"):
                # not a wildcard address, use it as-is:
                add_address(socktype, host, port)
                continue
            if not netifaces:
                if first_time("netifaces-socket-address"):
                    netlog("get_socket_info()", backtrace=True)
                    netlog.warn("Warning: netifaces is missing")
                    netlog.warn(" socket addresses cannot be queried")
            else:
                for ip in get_all_ips():
                    add_address(socktype, ip, port)

        for socktype, auth_classes in self.auth_classes.items():
            if auth_classes:
                authenticators = si.setdefault(socktype, {}).setdefault("authenticator", {})
                for i, auth_class in enumerate(auth_classes):
                    authenticators[i] = auth_class[0], auth_class[3]
        return si

    ######################################################################
    # packet handling:
    def process_packet(self, proto, packet) -> None:
        return super().dispatch_packet(proto, packet)

    def handle_rfb_connection(self, conn, data: bytes = b"") -> None:
        log.error("Error: RFB protocol is not supported by this server")
        log("handle_rfb_connection%s", (conn, data))
        force_close_connection(conn)

    def handle_rdp_connection(self, conn, data: bytes = b"") -> None:
        if data and data[:2] != b"\x03\x00":
            raise ValueError("packet is not valid RDP")
        log.error("Error: RDP protocol is not supported by this server")
        log("handle_rdp_connection%s", (conn, data))
        force_close_connection(conn)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.net.file_transfer import FileTransferHandler
from xpra.client.base.stub import StubClientMixin
from xpra.log import Logger

GLib = gi_import("GLib")

printlog = Logger("printing")
filelog = Logger("file")

DELETE_PRINTER_FILE = envbool("XPRA_DELETE_PRINTER_FILE", True)
SKIP_STOPPED_PRINTERS = envbool("XPRA_SKIP_STOPPED_PRINTERS", True)
INIT_PRINTING_DELAY = envint("XPRA_INIT_PRINTING_DELAY", 2)


class FileMixin(StubClientMixin, FileTransferHandler):

    def __init__(self):
        StubClientMixin.__init__(self)
        FileTransferHandler.__init__(self)
        self.remote_request_file: bool = False

    def init(self, opts) -> None:
        # printing and file transfer:
        FileTransferHandler.init_opts(self, opts)

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets(
            "open-url", "send-file",
            "send-data-request", "send-data-response",
            "ack-file-chunk", "send-file-chunk",
        )

    def get_caps(self) -> dict[str, Any]:
        return {"file": self.get_file_transfer_features()}

    def cleanup(self) -> None:
        # we must clean printing before FileTransferHandler, which turns the printing flag off!
        FileTransferHandler.cleanup(self)

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.parse_file_transfer_caps(c)
        self.remote_request_file = c.boolget("request-file", False)
        return True



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.net.glib_handler import GLibPacketHandler
from xpra.util.glib import register_os_signals, register_SIGUSR_signals
from xpra.server.subsystem.stub import StubServerMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("server", "glib")


class GLibServer(StubServerMixin, GLibPacketHandler):

    def __init__(self):
        StubServerMixin.__init__(self)
        GLibPacketHandler.__init__(self)
        self.main_loop = GLib.MainLoop()

    def __repr__(self):
        return "GLibServer"

    @staticmethod
    def install_signal_handlers(callback: Callable[[int], None]) -> None:
        sstr = "encoder server"
        register_os_signals(callback, sstr)
        register_SIGUSR_signals(sstr)

    def do_run(self) -> None:
        run: Callable = self.main_loop.run
        log("do_run() calling %s()", run)
        run()
        log("do_run() end of %()", run)

    def do_quit(self) -> None:
        log("do_quit: calling main_loop.quit()")
        self.main_loop.quit()
        # from now on, we can't rely on the main loop:
        from xpra.util.system import register_SIGUSR_signals
        register_SIGUSR_signals()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def get_backend_module() -> str:
    from xpra.platform.posix.gui import x11_bindings
    if x11_bindings():
        try:
            from xpra import x11
            assert x11
            return "xpra.x11.selection.clipboard.X11Clipboard"
        except ImportError:
            pass
    return "xpra.gtk.clipboard.GTK_Clipboard"



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def get_client_base_classes() -> tuple[type, ...]:
    from xpra.client.base import features
    from xpra.client.base.glib_adapter import GLibClient
    from xpra.client.base.serverinfo import ServerInfoMixin
    from xpra.client.base.network import NetworkClient
    from xpra.client.base.aes import AESClient
    from xpra.client.base.progress import ProgressClient
    from xpra.client.base.challenge import ChallengeClient
    CLIENT_BASES: list[type] = [GLibClient, NetworkClient, ServerInfoMixin, AESClient, ProgressClient, ChallengeClient]
    if features.debug:
        from xpra.client.base.debug import DebugClient
        CLIENT_BASES.append(DebugClient)
    if features.file:
        from xpra.client.base.file import FileMixin
        CLIENT_BASES.append(FileMixin)
    if features.printer:
        from xpra.client.base.printer import PrinterMixin
        CLIENT_BASES.append(PrinterMixin)
    if features.control:
        from xpra.client.base.control import ControlClient
        CLIENT_BASES.append(ControlClient)
    from xpra.net.common import SSL_UPGRADE
    if SSL_UPGRADE:
        from xpra.client.base.ssl_upgrade import SSLUpgradeClient
        CLIENT_BASES.append(SSLUpgradeClient)
    return tuple(CLIENT_BASES)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.net.common import Packet
from xpra.server.subsystem.stub import StubServerMixin
from xpra.x11.bindings.core import get_root_xid
from xpra.x11.dispatch import add_event_receiver
from xpra.x11.common import X11Event
from xpra.log import Logger

log = Logger("bell")


class BellServer(StubServerMixin):
    """
    Servers that forward bell events.
    """
    PREFIX = "bell"
    __signals__ = {
        "x11-xkb-event": 1,
    }

    def __init__(self):
        StubServerMixin.__init__(self)
        self.bell = False

    def init(self, opts) -> None:
        self.bell = opts.bell
        log(f"bell={opts.bell}")

    def get_caps(self, source) -> dict[str, Any]:
        # Note: don't just call self.get_info() to get rid of linter warnings,
        # this is not safe as it will call it on the subclass!
        return {
            "bell": self.bell,
        }

    def get_info(self, _proto) -> dict[str, Any]:
        return {
            "bell": self.bell,
        }

    def setup(self) -> None:
        if not self.bell:
            return
        try:
            from xpra.x11.bindings.keyboard import X11KeyboardBindings
            X11Keyboard = X11KeyboardBindings()
        except ImportError as e:
            log.error("Error: unable to listen for bell events:")
            log.estr(e)
            self.bell = False
        else:
            from xpra.x11.error import xlog
            with xlog:
                X11Keyboard.selectBellNotification(True)
                # somewhat redundant, but doing it more than once does not hurt:
                rxid = get_root_xid()
                add_event_receiver(rxid, self)

    def _process_bell_set(self, proto, packet: Packet) -> None:
        assert self.bell, "cannot toggle send_bell: the feature is disabled"
        ss = self.get_server_source(proto)
        if ss:
            ss.send_bell = packet.get_bool(1)

    def do_x11_xkb_event(self, event: X11Event) -> None:
        log("server do_x11_xkb_event(%r) bell=%s", event, self.bell)
        if not self.bell:
            return
        # X11: XKBNotify
        if event.subtype != "bell":
            log.error(f"Error: unknown event subtype: {event.subtype!r}")
            log.error(f" {event=}")
            return
        # bell events on our windows will come through the bell signal,
        # this method is a catch-all for events on windows we don't manage,
        # so we use wid=0 for that:
        wid = 0
        for ss in self.window_sources():
            ss.bell(wid, event.device, event.percent,
                    event.pitch, event.duration, event.bell_class, event.bell_id, event.name)

    def _bell_signaled(self, wm, event) -> None:
        log("bell signaled on window %#x", event.window)
        if not self.bell:
            return
        wid = 0
        rxid = get_root_xid()
        if event.window != rxid and event.window_model is not None:
            wid = self._window_to_id.get(event.window_model, 0)
        log("_bell_signaled(%s,%r) wid=%#x", wm, event, wid)
        for ss in self.window_sources():
            ss.bell(wid, event.device, event.percent,
                    event.pitch, event.duration, event.bell_class, event.bell_id, event.name)

    def init_packet_handlers(self) -> None:
        if self.bell:
            self.add_packets("bell-set", main_thread=True)
            self.add_legacy_alias("set-bell", "bell-set")



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence

from xpra.net.compression import Compressed
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger

log = Logger("encoding")


def make_screenshot_packet_from_regions(regions: Sequence[tuple[int, int, int, ImageWrapper]]) \
        -> tuple[str, int, int, str, int, Any]:
    # regions = array of (wid, x, y, PIL.Image)
    if not regions:
        log("screenshot: no regions found, returning empty 0x0 image!")
        return "screenshot", 0, 0, "png", 0, b""
    # in theory, we could run the rest in a non-UI thread since we're done with GTK..
    minx: int = min(x for (_, x, _, _) in regions)
    miny: int = min(y for (_, _, y, _) in regions)
    maxx: int = max((x + img.get_width()) for (_, x, _, img) in regions)
    maxy: int = max((y + img.get_height()) for (_, _, y, img) in regions)
    width = maxx - minx
    height = maxy - miny
    log("screenshot: %sx%s, min x=%s y=%s", width, height, minx, miny)
    from PIL import Image  # pylint: disable=import-outside-toplevel
    screenshot = Image.new("RGBA", (width, height))
    for wid, x, y, img in reversed(regions):
        pixel_format = img.get_pixel_format()
        target_format = {
            "XRGB": "RGB",
            "BGRX": "RGB",
            "BGRA": "RGBA",
        }.get(pixel_format, pixel_format)
        pixels = img.get_pixels()
        w = img.get_width()
        h = img.get_height()
        # PIL cannot use the memoryview directly:
        if isinstance(pixels, memoryview):
            pixels = pixels.tobytes()
        try:
            window_image = Image.frombuffer(target_format, (w, h), pixels, "raw", pixel_format, img.get_rowstride())
        except (ValueError, TypeError):
            log.error("Error parsing window pixels in %s format for window %i", pixel_format, wid, exc_info=True)
            continue
        tx = x - minx
        ty = y - miny
        screenshot.paste(window_image, (tx, ty))
    from io import BytesIO
    buf = BytesIO()
    screenshot.save(buf, "png")
    data = buf.getvalue()
    buf.close()
    compressed = Compressed("png", data)
    packet = ("screenshot", width, height, "png", width * 4, compressed)
    log("screenshot: %sx%s %s", width, height, compressed)
    return packet
